from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
# ----------------------------------------------------------------------
# clex.py
#
# A lexer for ANSI C.
# ----------------------------------------------------------------------

import sys
sys.path.insert(0, "../..")

try:
    import pymel.util.external.ply.lex as lex
except ImportError:
    import ply.lex as lex

# Reserved words

# removed 'AUTO', 'CONST',  'CHAR', 'DOUBLE','ENUM', 'EXTERN', 'LONG', 'REGISTER', 'SHORT', 'SIGNED', 'STATIC', 'STRUCT', 'TYPEDEF','UNION', 'UNSIGNED', 'VOID', 'VOLATILE','GOTO',


reserved = (
    'BREAK', 'CASE', 'CONTINUE', 'DEFAULT', 'DO',
    'ELSE', 'FALSE', 'FLOAT', 'FOR', 'GLOBAL', 'IF', 'IN', 'INT', 'NO', 'ON', 'OFF', 'PROC',
    'RETURN', 'STRING', 'SWITCH', 'TRUE', 'VECTOR', 'MATRIX',
    'WHILE', 'YES',
)

tokens = reserved + (
    # Literals (identifier, integer constant, float constant, string constant, char const)
    'ID', 'VAR', 'ICONST', 'FCONST', 'SCONST',
    #'LOBJECT', 'ROBJECT',

    # Operators (+,-,*,/,%,^,<<,>>, ||, &&, !, <, <=, >, >=, ==, !=)
    'PLUS', 'MINUS', 'TIMES', 'DIVIDE', 'MOD',
    'NOT', 'CROSS',
    'LOR', 'LAND',
    'LT', 'LE', 'GT', 'GE', 'EQ', 'NE',

    # Assignment (=, *=, /=, %=, +=, -=, ^=)
    'EQUALS', 'TIMESEQUAL', 'DIVEQUAL', 'MODEQUAL', 'PLUSEQUAL', 'MINUSEQUAL',
    'CROSSEQUAL',

    # Vector Component
    'COMPONENT',

    # Increment/decrement (++,--)
    'PLUSPLUS', 'MINUSMINUS',

    # Conditional operator (?)
    'CONDOP',

    # Delimeters ( ) [ ] { } , . ; :
    'LPAREN', 'RPAREN',
    'LBRACKET', 'RBRACKET',
    'LBRACE', 'RBRACE',
    'COMMA', 'SEMI', 'COLON',
    'CAPTURE',
    'LVEC', 'RVEC',

    # Comments
    'COMMENT', 'COMMENT_BLOCK',

    # Ellipsis (..)
    'ELLIPSIS',
)

# Completely ignored characters
t_ignore = ' \t\x0c'

# Newlines


def t_NEWLINE(t):
    r'\n+|\r+'
    t.lexer.lineno += t.value.count("\n")


# Operators
t_PLUS = r'\+'
t_MINUS = r'-'
t_TIMES = r'\*'
t_DIVIDE = r'/'
t_MOD = r'%'
#t_OR               = r'\|'
#t_AND               = r'&'
t_NOT = r'!'
t_CROSS = r'\^'
t_LVEC = r'<<'
t_RVEC = r'>>'
t_LOR = r'\|\|'
t_LAND = r'&&'
t_LT = r'<'
t_GT = r'>'
t_LE = r'<='
t_GE = r'>='
t_EQ = r'=='
t_NE = r'!='

# Assignment operators

t_EQUALS = r'='
t_TIMESEQUAL = r'\*='
t_DIVEQUAL = r'/='
t_MODEQUAL = r'%='
t_PLUSEQUAL = r'\+='
t_MINUSEQUAL = r'-='
t_CROSSEQUAL = r'^='

# Increment/decrement
t_PLUSPLUS = r'\+\+'
t_MINUSMINUS = r'--'

# ?
t_CONDOP = r'\?'

# Delimeters
#t_LBRACKET           = r'\['
#t_RBRACKET           = r'\]'
t_LBRACE = r'\{'
t_RBRACE = r'\}'
t_COMMA = r','
#t_SEMI               = r';'
t_COLON = r':'
#t_COMPONENT           = r'\.[xyz]'

# Identifiers and reserved words

reserved_map = {}
for r in reserved:
    reserved_map[r.lower()] = r

# print reserved_map

id_state = None
suspend_depth = 0


def t_LPAREN(t):
    r'\('
    return t


def t_RPAREN(t):
    r'\)'
    return t

# def t_LOBJECT(t):
#    r'([|]?([:]?([.]?[A-Za-z_][\w]*)+)+)+?\['
#    return t

# def t_ROBJECT(t):
#    r'\]([|]?([:]?([.]?[A-Za-z_][\w]*)+)+)+?'
#    return t


def t_LBRACKET(t):
    r'\['
    return t


def t_RBRACKET(t):
    r'\]'
    return t


def t_CAPTURE(t):
    r'`'
    return t


def t_SEMI(t):
    r';'
    return t


def t_VAR(t):
    r'\$[A-Za-z_][\w_]*'
    return t


def t_COMPONENT(t):
    r'\.[xyz]'
    return t


def t_ELLIPSIS(t):
    r'\.\.'
    return t


def t_ID(t):
    # Starts with a letter or a pipe
    #
    # |path|myPrfx_1:myNode_1.myAttr_1[0].subAttr   or  ..
    # r'[A-Za-z_|](?:[\w_\.:|]|(?:\[\d+\]))*|\.\.'
    # r'[A-Za-z_|]([\w_\.:|]|(\[\d+\]))*|\.\.'
    # r'[A-Za-z_|][\w_\.:|]*(?:[\w_]|(?:\[\d+\]))+|[A-Za-z_]|\.\.'
    # '(%(id)s)*([.|](%(id)s)+)*[.|]?*|\.\.' % {'id': '[A-Za-z_][\w:]*(\[\d+\])?'}

    #   |    :    .     idName        [0]
    # r'([|]?([:]?([.]?[A-Za-z_][\w]*(\[\d+\])?)+)+)+?|\.\.'
    #   |    :    .     idName        [0] or [$var]
    # r'([|]?([:]?([.]?[A-Za-z_][\w]*(\[(\d+)|(\$[A-Za-z_][\w_]*)\])?)+)+)+?|\.\.'

    # r'[A-Za-z_][\w]*'
    r'([|]?([:]?([.]?[A-Za-z_][\w]*)+)+)+?'
    t.type = reserved_map.get(t.value, "ID")
    return t

# def t_OBJECT(t):
#
#    #   |    :    .     idName        [0]
#    #r'([|]?([:]?([.]?[A-Za-z_][\w]*(\[\d+\])?)+)+)+?|\.\.'
#
#    #r'[A-Za-z_][\w]|\.\.'
#    r'([|]?([:]?([.]?[A-Za-z_][\w]*)+)+)+?'
#    return t

# Integer literal
#t_ICONST = r'\d+([uU]|[lL]|[uU][lL]|[lL][uU])?'
t_ICONST = r'(0x[a-fA-F0-9]*)|\d+'

# Floating literal
# t_FCONST = r'((\d+)?(\.\d+)(e(\+|-)?(\d+))?|(\d+)e(\+|-)?(\d+))([lL]|[fF])?' # does not allow  1.
t_FCONST = r'(((\d+\.)(\d+)?|(\d+)?(\.\d+))(e(\+|-)?(\d+))?|(\d+)e(\+|-)?(\d+))([lL]|[fF])?'

# String literal
# t_SCONST = r'\"([^\\\n]|(\\.))*?\"' # does not allow \ for spanning multiple lines
#t_SCONST = '\"([^\n]|\r)*\"'
#t_SCONST = r'"([^\n]|\\\n)*?"'
t_SCONST = r'"([^\\\n]|(\\.)|\\\n)*?"'

# Comments


def t_COMMENT_BLOCK(t):
    r'/\*(.|\n)*?\*/|/\*(.|\n)*?$'
    # r'/\*(.|\n)*?\*/'

    # r'/\*(.|\n)*?\*/|/\*(.|\n)*?$'
    # the second half of this regex is for matching block comments that
    # are terminated by the end of the file instead of by */

    t.lexer.lineno += t.value.count('\n')
    return t


def t_COMMENT(t):
    r'//.*'
    #t.lexer.lineno += t.value.count('\n')
    return t

# def t_INVALID(t):
#    r'[.~\|\'"]'
#    # these symbols cannot be used on their own,
#    # without this entry, they would simply be ignored rather than raising an error
#    return t

# def t_error(t):
#    #print "Illegal character %s" % repr(t.value[0])
#    t.lexer.skip(1)

# def t_error(t):
#    return t

#lexer = lex.lex(optimize=1)
#lexer = lex.lex()
# if __name__ == "__main__":
#    lex.runmain(lexer)



"""Imports all of pymel into one namespace, for use during interactive scripting"""
from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import sys
import pymel as _pymel
_pymel.all = sys.modules[__name__]

from pymel import internal
import pymel.internal.startup
doFinalize = pymel.internal.startup.finalizeEnabled
pymel.internal.startup.finalizeEnabled = False
from pymel.internal import plogging
from pymel.internal import factories
from pymel import mayautils

#logger = plogging.getplogging.pymelLogger(__name__)
plogging.pymelLogger.debug('imported internal')

from pymel import api
plogging.pymelLogger.debug('imported api')

from pymel.core import *
plogging.pymelLogger.debug('imported core')

# for wrapped math functions
from pymel.util.arrays import *

from pymel.core import datatypes

from pymel import versions

from pymel.core import nodetypes
from pymel.core.nodetypes import *

# if in batch mode we may not have UI commands
if not cmds.about(batch=True):
    from pymel.core.uitypes import *

# These two were imported into 'old' pymel top level module,
# so make sure they're imported here as well
from pymel import core
from pymel import tools

# some submodules do 'import pymel.core.pmcmds as cmds' -
# this ensures that when the user does 'from pymel import *',
# cmds is always maya.cmds
import maya.cmds as cmds

# Run delayed finalize now, so that if userSetup imports all,
# it has access to everything it should
pymel.internal.startup.finalizeEnabled = doFinalize
pymel.internal.startup.finalize()



from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
#------------------------------------------------------------------------------
# Condition objects - used for chaining together tests that yield True/False results
#------------------------------------------------------------------------------


from builtins import object
class NO_DATA(Exception):
    pass


class Condition(object):

    """
    Used to chain together objects for conditional testing.
    """

    def __init__(self, value=None):
        self.value = value

    def eval(self, data=NO_DATA):
        return bool(self.value)

    def __or__(self, other):
        return Or(self, other)

    def __ror__(self, other):
        return Or(other, self)

    def __and__(self, other):
        return And(self, other)

    def __rand__(self, other):
        return And(other, self)

    def __invert__(self):
        return Inverse(self)

    def __bool__(self):
        return self.eval()

    def __str__(self):
        return str(self.value)

Always = Condition(True)

Never = Condition(False)


class Inverse(Condition):

    def __init__(self, toInvert):
        self.toInvert = toInvert

    def eval(self, data=NO_DATA):
        return not self.toInvert.eval(data)

    def __str__(self):
        return "not %s" % self.toInvert


class AndOrAbstract(Condition):

    def __init__(self, *args):
        self.args = []
        for arg in args:
            if isinstance(arg, self.__class__):
                self.args.extend(arg.args)
            else:
                self.args.append(arg)

    def eval(self, data=NO_DATA):
        for arg in self.args:
            if isinstance(arg, Condition):
                val = arg.eval(data)
            else:
                val = bool(arg)
            if val == self._breakEarly:
                return self._breakEarly
        return not self._breakEarly

    def __str__(self):
        return "(%s)" % self._strJoiner.join([str(x) for x in self.args])


class And(AndOrAbstract):
    _breakEarly = False
    _strJoiner = ' and '


class Or(AndOrAbstract):
    _breakEarly = True
    _strJoiner = ' or '



"""
Redefine format warning to avoid getting garbage at end of line when raised directly from Maya console
and define a UserWarning class that does only print it's message (no line or module info)


"""
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from future.utils import PY2
from past.builtins import basestring
import warnings


def formatwarning(message, category, filename, lineno, line=None):
    """Redefined format warning for maya."""
    if issubclass(category, ExecutionWarning):
        s = u"%s: %s\n" % (category.__name__, message)
    else:
        s = u'%s: %s, at line %s, in "%s"\n' % (category.__name__, message, lineno, filename)
#        name, ext = os.path.splitext(filename)
#        line = ""
#        if ext == ".py" :
#            line = unicode(linecache.getline(filename, lineno)).strip()
#            if line:
#                s += (u"#\t %s" % line)
    return s

warnings.formatwarning = formatwarning

# def showwarning(message, category, filename, lineno, file=None, line=None):
#    msg = warnings.formatwarning(message, category, filename, lineno, line)
#    if file:
#        msg += " >> %r" % file
#    _logger.warning(msg)
#
#warnings.showwarning = showwarning

# Subclass just to allow users to configure filtering of pymel-specific
# deprecations
class PymelBaseWarning(Warning):
    pass

class ExecutionWarning (UserWarning, PymelBaseWarning):

    """ Simple Warning class that doesn't print any information besides warning message """

class PymelBaseDeprecationWarning(PymelBaseWarning):
    pass

# Subclass from FutureWarning so it's displayed by default
class PymelFutureWarning(FutureWarning, PymelBaseDeprecationWarning):
    pass

# Subclass from DeprecationWarning so it's not displayed by default
class MayaDeprecationWarning(DeprecationWarning, PymelBaseDeprecationWarning):
    pass


def warn(*args, **kwargs):
    """ Default Maya warn which uses ExecutionWarning as the default warning class. """
    if len(args) == 1 and not isinstance(args[0], Warning):
        args = args + (ExecutionWarning,)
    stacklevel = kwargs.pop("stacklevel", 1) + 1  # add to the stack-level so that this wrapper func is skipped
    return warnings.warn(stacklevel=stacklevel, *args, **kwargs)


def deprecated(funcOrMessage=None, className=None,
               baseMessage="The function '{objName}' is deprecated and will"
                           " become unavailable in future pymel versions",
               warningType=PymelFutureWarning):
    """Decorates a function so that it prints a deprecation warning when called.

    The decorator can either receive parameters or the function directly.

    Parameters
    ----------
    funcOrMessage : Union[str, Callable[..., Any], None]
        If passed a message, the message will be appended to the standard
        deprecation warning and should serve to further clarify why the function
        is being deprecated and/or suggest an alternative function. In this
        case, the return result of this function is another decorator (with the
        ammended message), which then needs to be fed the function to be
        decorated. Otherwise, funcOrMessage should be the func to be decorated,
        and the return result is decorated version of funcOrMessage
    className : Union[str, False, None]
        If given as a str, then the decorated function is asssumed to be method,
        and the name is printed as "module.className.funcName".  If False, it
        is assumed to NOT be a method, and the name is printed as
        "module.funcName".  If None, then the decorator will try to
        automatically determine whether the passed function is a method, and if
        so, what it's className is.
    baseMessage : Optional[str]
        Message which will be combined with the optional message (in
        funcOrMessage) to form the final message. Maybe set to None to ensure
        only the message (in funcOrMessage) is printed.
    warningType : Type[Warning]
        Warning class to raise. Note that DeprecationWarning is ignored by
        default.
    """
    import inspect

    def isClassMethodOrMethod(test_func):
        isClassMethod = False
        isMethod = False
        args = list(inspect.signature(test_func).parameters)
        if args:
            if args[0] == 'cls':
                isClassMethod = True
            elif args[0] == 'self':
                isMethod = True
        return isClassMethod, isMethod

    if PY2:
        def isClassMethodOrMethod(test_func):
            isClassMethod = False
            isMethod = False
            args = inspect.getargspec(test_func).args
            if args:
                if args[0] == 'cls':
                    isClassMethod = True
                elif args[0] == 'self':
                    isMethod = True
            return isClassMethod, isMethod

    #@decorator
    def deprecated2(func):
        useClassName = False
        info = dict(
            name=func.__name__,
            module=func.__module__)

        if className is None:
            isClassMethod, isMethod =isClassMethodOrMethod(func)
            if isClassMethod or isMethod:
                useClassName = True
        elif className is not False:
            useClassName = True
            info['className'] = className

        if useClassName:
            objName = '%(module)s.%(className)s.%(name)s'
        else:
            objName = '%(module)s.%(name)s'
        message2 = message.format(objName=objName)

        def deprecationLoggedFunc(*args, **kwargs):
            if useClassName and className is None:
                if isClassMethod:
                    info['className'] = args[0].__name__
                else:
                    info['className'] = type(args[0]).__name__
            # add to the stack-level so that this wrapper func is skipped
            warnings.warn(message2 % info, warningType, stacklevel=2)
            return func(*args, **kwargs)

        deprecationLoggedFunc.__name__ = func.__name__
        deprecationLoggedFunc.__module__ = func.__module__
        deprecationLoggedFunc.__doc__ = message % info
        deprecationLoggedFunc._func_before_deprecation = func
        if func.__doc__:
            deprecationLoggedFunc.__doc__ += '\n\n' + func.__doc__
        return deprecationLoggedFunc

    # check if the decorator got a 'message' parameter
    if funcOrMessage is None:
        message = baseMessage
        return deprecated2
    elif isinstance(funcOrMessage, basestring):
        if baseMessage is None:
            message = funcOrMessage
        else:
            message = baseMessage + '. ' + funcOrMessage
        return deprecated2
    else:
        message = baseMessage
        return deprecated2(funcOrMessage)


def maya_deprecated(
        funcOrMessage=None, className=None,
        baseMessage="The function '{objName}' has been deprecated by maya and"
                    " may become unavailable in future maya versions",
        warningType=MayaDeprecationWarning):
    return deprecated(funcOrMessage=funcOrMessage, className=className,
                      baseMessage=baseMessage, warningType=warningType)

if __name__ == '__main__':
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-

# enum.py
# Part of enum, a package providing enumerated types for Python.
#
# Copyright © 2007 Ben Finney
# This is free software; you may copy, modify and/or distribute this work
# under the terms of the GNU General Public License, version 2 or later
# or, at your option, the terms of the Python license.

"""Robust enumerated type support in Python

This package provides a module for robust enumerations in Python.

An enumeration object is created with a sequence of string arguments
to the Enum() constructor:

    >>> from .enum import Enum
    >>> Colours = Enum('Colours', ['red', 'blue', 'green'])
    >>> Weekdays = Enum('Weekdays', ['mon', 'tue', 'wed', 'thu', 'fri', 'sat', 'sun'])

The return value is an immutable sequence object with a value for each
of the string arguments. Each value is also available as an attribute
named from the corresponding string argument:

    >>> pizza_night = Weekdays[4]
    >>> shirt_colour = Colours.green

The values are constants that can be compared with values from
the same enumeration, as well as with integers or strings; comparison with other
values will invoke Python's fallback comparisons:

    >>> pizza_night == Weekdays.fri
    True
    >>> shirt_colour > Colours.red
    True
    >>> shirt_colour == "green"
    True

Each value from an enumeration exports its sequence index
as an integer, and can be coerced to a simple string matching the
original arguments used to create the enumeration:

    >>> str(pizza_night)
    'fri'
    >>> shirt_colour.index
    2
"""
from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

from past.builtins import cmp
from future import standard_library

# 2to3: remove switch when python-3 only
try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping
standard_library.install_aliases()
from past.builtins import basestring
from builtins import object
__author_name__ = "Ben Finney"
__author_email__ = "ben+python@benfinney.id.au"
#__author__ = "%s <%s>" % (__author_name__, __author_email__) # confuses epydoc
__date__ = "2007-01-24"
__copyright__ = "Copyright © %s %s" % (
    __date__.split('-')[0], __author_name__
)
__license__ = "Choice of GPL or Python license"
__url__ = "http://cheeseshop.python.org/pypi/enum/"
__version__ = "0.4.3"

import operator
from collections import OrderedDict


class EnumException(Exception):

    """ Base class for all exceptions in this module """

    def __init__(self):
        if self.__class__ is EnumException:
            raise NotImplementedError("%s is an abstract class for subclassing" % self.__class__)


class EnumEmptyError(AssertionError, EnumException):

    """ Raised when attempting to create an empty enumeration """

    def __str__(self):
        return "Enumerations cannot be empty"


class EnumBadKeyError(TypeError, EnumException):

    """ Raised when creating an Enum with non-string keys """

    def __init__(self, key):
        self.key = key

    def __str__(self):
        return "Enumeration keys must be strings: %r" % (self.key,)


class EnumImmutableError(TypeError, EnumException):

    """ Raised when attempting to modify an Enum """

    def __init__(self, *args):
        self.args = args

    def __str__(self):
        return "Enumeration does not allow modification"


class EnumBadDefaultKeyError(ValueError, EnumException):

    """ Raised when a supplied default key for a value was not present """

    def __init__(self, val, key):
        self.val = val
        self.key = key

    def __str__(self):
        return "Given default key %r for index %r not present in keys" % (self.key, self.val)


class EnumValue(object):

    """ A specific value of an enumerated type """

    def __init__(self, enumtype, index, key, doc=None):
        """ Set up a new instance """
        self.__enumtype = enumtype
        self.__index = index
        self.__key = key
        self.__doc = doc

    def __get_enumtype(self):
        return self.__enumtype
    enumtype = property(__get_enumtype)

    def __get_key(self):
        return self.__key
    key = property(__get_key)

    def __str__(self):
        return "%s" % (self.key)

    def __int__(self):
        return self.index

    def __get_index(self):
        return self.__index
    index = property(__get_index)

    def __repr__(self):
        if self.__doc:
            return "EnumValue(%r, %r, %r, %r, %r)" % (
                self.__enumtype._name,
                self.__index,
                self.__key,
                self.__doc,
            )
        else:
            return "EnumValue(%r, %r, %r)" % (
                self.__enumtype._name,
                self.__index,
                self.__key,
            )

    def _asTuple(self):
        return (self.__index, self.__key, self.__doc)

    def __hash__(self):
        return hash(self.__index)

    def _get_comparers(self, other):
        result = NotImplemented
        if isinstance(other, EnumValue) and self.enumtype == other.enumtype:
            result = (self.index, other.index)
        elif isinstance(other, basestring):
            result = (self.key, other)
        elif isinstance(other, int):
            result = (self.index, other)
        return result

    def __eq__(self, other):
        comparers = self._get_comparers(other)
        if comparers is NotImplemented:
            return NotImplemented
        return comparers[0] == comparers[1]

    def __ne__(self, other):
        comparers = self._get_comparers(other)
        if comparers is NotImplemented:
            return NotImplemented
        return comparers[0] != comparers[1]

    def __gt__(self, other):
        comparers = self._get_comparers(other)
        if comparers is NotImplemented:
            return NotImplemented
        return comparers[0] > comparers[1]

    def __lt__(self, other):
        comparers = self._get_comparers(other)
        if comparers is NotImplemented:
            return NotImplemented
        return comparers[0] < comparers[1]

    def __ge__(self, other):
        comparers = self._get_comparers(other)
        if comparers is NotImplemented:
            return NotImplemented
        return comparers[0] >= comparers[1]

    def __le__(self, other):
        comparers = self._get_comparers(other)
        if comparers is NotImplemented:
            return NotImplemented
        return comparers[0] <= comparers[1]

# Modified to support multiple keys for the same value


class Enum(object):

    """ Enumerated type """

    def __init__(self, name, keys, **kwargs):
        # type: (str, Union[Dict[str, int], Iterable[str], Iterable[Tuple[str, int]]], **Any) -> None
        """ Create an enumeration instance

        Parameters
        ----------
        name : str
            The name of this enumeration
        keys : Union[Dict[str, int], Iterable[str], Iterable[Tuple[str, int]]]
            The keys for the enumeration; if this is a dict, it should map
            from key to it's value (ie, from string to int)
            Otherwise, it should be an iterable of keys, where their index
            within the iterable is their value, or an iterable of key/value
            pairs - ie, passing any of these would give the same result:
                {'Red':0,'Green':1,'Blue':2}
                ('Red', 'Green', 'Blue')
                [('Green', 1), ('Blue', 2), ('Red', 0)]
        multiKeys : bool
            Defaults to False
            If True, allows multiple keys per value - ie,
                Enum('Names', {'Bob':0,'Charles':1,'Chuck':1}, multiKeys=True)
            When looking up a key from a value, a single key is always returned
            - see defaultKeys for a discussion of which key this is.
            When multiKeys is enabled, the length of keys and values may not be
            equal.
            If False (default), then the end result enum will always have a
            one-to-one key / value mapping; if multiple keys are supplied for a
            a single value, then which key is used is indeterminate (an error
            will not be raised).
        defaultKeys : Dict[int, str]
            If given, should be a map from values to the 'default' key to
            return for that value when using methods like getKey(index)
            This will only be used if the value actually has multiple keys
            mapping to it, and in this case, the specified default key must be
            present within keys (if not, a EnumBadDefaultKeyError is raised).
            If there are multiple keys for a given value, and no defaultKey is
            provided, the default key is whichever one is encountered first
            when iterating through the `key` parameter (which is well-defined
            for a list of pairs or an OrderedDict, and effectively random for
            a normal dict)
        docs : Optional[Dict[int, str]]
            if given, should provide a map from keys to an associated docstring
            for that key; the dict need not provide an entry for every key
        """
        if not keys:
            raise EnumEmptyError()

        defaultKeys = kwargs.pop('defaultKeys', {})
        multiKeys = kwargs.pop('multiKeys', False)
        docs = kwargs.pop('docs', {})

        # Keys for which there are multiple keys mapping to the same
        # value, but are not the default key for that value
        extraKeys = {}
        pairs = None

        if isinstance(keys, Mapping):
            pairs = list(keys.items())
        else:
            # could be an iterable of keys - in which case the "value" is it's
            # index - or it could be an iterable of pairs, (key, value).
            # Unfortunately, we don't know which until we start to iterate...
            # ...and since keys may be an iterator, where iterating is
            # "destructive", to simplify we just convert it to a new list
            keys = list(keys)
            if not keys:
                # an iterable would evauluate to a True boolean even if empty -
                # so redo our empty test
                raise EnumEmptyError()
            if isinstance(keys[0], basestring):
                keygen = enumerate(keys)
                values = [None] * len(keys)
            else:
                pairs = keys

        if pairs is not None:
            if not multiKeys:
                reverse = dict([(v, k) for k, v in pairs])
            else:
                reverse = dict()
                for key, val in pairs:
                    reverse.setdefault(val, []).append(key)
                for val, keyList in reverse.items():
                    if len(keyList) == 1:
                        defaultKey = keyList[0]
                    else:
                        if val in defaultKeys:
                            defaultKey = defaultKeys[val]
                            if defaultKey not in keyList:
                                raise EnumBadDefaultKeyError(val, defaultKey)
                        else:
                            defaultKey = keyList[0]
                        for multiKey in keyList:
                            if multiKey != defaultKey:
                                extraKeys[multiKey] = val
                    reverse[val] = defaultKey
            keygen = [(v, reverse[v]) for v in sorted(reverse.keys())]
            values = {}

        value_type = kwargs.get('value_type', EnumValue)

        keyDict = {}
        for val, key in keygen:
            # print val, key
            value = value_type(self, val, key, docs.get(key))
            values[val] = value
            keyDict[key] = val
            try:
                super(Enum, self).__setattr__(key, value)
            except TypeError as e:
                raise EnumBadKeyError(key)

        for key, val in extraKeys.items():
            # throw away any docs for the extra keys
            keyDict[key] = val

        # always store values as an OrderedDict, to provide unified behavior,
        # regardless of how it's constructed
        if not isinstance(values, Mapping):
            values = OrderedDict(enumerate(values))
        elif not isinstance(values, OrderedDict):
            values = OrderedDict((key, values[key]) for key in sorted(values))

        super(Enum, self).__setattr__('_keys', keyDict)
        super(Enum, self).__setattr__('_values', values)
        super(Enum, self).__setattr__('_docs', docs)
        super(Enum, self).__setattr__('_name', name)

    @property
    def name(self):
        return self._name

    def _valTuples(self):
        return tuple((key, val._asTuple())
            for (key, val) in self._values.items())

    def __eq__(self, other):
        if not isinstance(other, Enum):
            return False
        if not self._keys == other._keys:
            return False
        # For values, can't just compare them straight up, as the values
        # contain a ref to this Enum class, and THEIR compare compares the
        # Enum class - which would result in a recursive cycle
        # Instead, compare the values' _asTuple
        return self._valTuples() == other._valTuples()

    def __hash__(self):
        keyTuples = tuple((k, self._keys[k]) for k in sorted(self._keys))
        return hash((keyTuples, self._valTuples()))

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        if len(self._keys) != len(self._values):
            # multiKeys
            indexDict = {}
            defaults = {}
            for key, index in self._keys.items():
                keysForIndex = indexDict.setdefault(index, [])
                if len(keysForIndex) == 1:
                    defaults[index] = self._values[index].key
                keysForIndex.append(key)
            keysFlat = []

            # within each index, sort so that default value is always first,
            # and other keys are sorted alphabetically
            def sortKey(enumName):
                return (0 if enumName == default else 1, enumName)

            for enumValue in self.values():
                default = defaults.get(enumValue.index)
                for key in sorted(indexDict[enumValue.index], key=sortKey):
                    keysFlat.append((key, enumValue.index))
            keysFlat = ['({!r}, {!r})'.format(key, index)
                        for key, index in keysFlat]
            keysRepr = '[{}], multiKeys=True'.format(', '.join(keysFlat))
        else:
            keysFlat = [(ev.key, ev.index) for ev in self.values()]
            keysFlat = ['{!r}: {!r}'.format(key, index)
                        for key, index in keysFlat]
            keysRepr = '{' + ', '.join(keysFlat) + '}'

        return '{}({!r}, {})'.format(
            self.__class__.__name__, self.name, keysRepr)

    def __str__(self):
        return '%s%s' % (self.__class__.__name__, list(self.keys()))

    def __setattr__(self, name, value):
        raise EnumImmutableError(name)

    def __delattr__(self, name):
        raise EnumImmutableError(name)

    def __len__(self):
        return len(self._values)

    def __getitem__(self, index):
        return self._values[index]

    def __setitem__(self, index, value):
        raise EnumImmutableError(index)

    def __delitem__(self, index):
        raise EnumImmutableError(index)

    def __iter__(self):
        return iter(self._values)

    def __contains__(self, value):
        is_member = False
        if isinstance(value, basestring):
            is_member = (value in self._keys)
        else:
            # EnumValueCompareError was never defined...
            #            try:
            #                is_member = (value in self._values)
            #            except EnumValueCompareError:
            #                is_member = False
            is_member = (value in self._values)
        return is_member

    def getIndex(self, key):
        """Get an index value from a key
        This method always returns an index. If a valid index is passed instead
        of a key, the index will be returned unchanged.  This is useful when you
        need an index, but are not certain whether you are starting with a key
        or an index.

            >>> units = Enum('units', ['invalid', 'inches', 'feet', 'yards', 'miles', 'millimeters', 'centimeters', 'kilometers', 'meters'])
            >>> units.getIndex('inches')
            1
            >>> units.getIndex(3)
            3
            >>> units.getIndex('hectares')
            Traceback (most recent call last):
              ...
            ValueError: invalid enumerator key: 'hectares'
            >>> units.getIndex(10)
            Traceback (most recent call last):
              ...
            ValueError: invalid enumerator index: 10
        """
        if isinstance(key, int):
            # got a potential index : checking if it's valid
            if key in self._values:
                return key
            else:
                raise ValueError("invalid enumerator index: %r" % (key,))
        else:
            # got a key: retrieving index
            try:
                return self._keys[str(key)]
            except:
                raise ValueError("invalid enumerator key: %r" % (key,))

    def getKey(self, index):
        """
        Get a key value from an index
        This method always returns a key. If a valid key is passed instead of an
        index, the key will be returned unchanged.  This is useful when you need
        a key, but are not certain whether you are starting with a key or an
        index.

            >>> units = Enum('units', ['invalid', 'inches', 'feet', 'yards', 'miles', 'millimeters', 'centimeters', 'kilometers', 'meters'])
            >>> units.getKey(2)
            'feet'
            >>> units.getKey('inches')
            'inches'
            >>> units.getKey(10)
            Traceback (most recent call last):
              ...
            ValueError: invalid enumerator index: 10
            >>> units.getKey('hectares')
            Traceback (most recent call last):
              ...
            ValueError: invalid enumerator key: 'hectares'
        """

        if isinstance(index, int):
            # got an index: retrieving key
            try:
                return self._values[index].key
            except:
                raise ValueError("invalid enumerator index: %r" % index)
        else:
            # got a potential key : checking if it's valid
            if str(index) in self._keys:
                return index
            else:
                raise ValueError("invalid enumerator key: %r" % index)

    def values(self):
        "return a list of `EnumValue`s"
        return tuple(self._values.values())

    def itervalues(self):
        "iterator over EnumValue objects"
        return iter(self._values.values())

    def keys(self):
        "return a list of keys as strings"
        if not hasattr(self, '_keyStrings'):
            keyStrings = tuple(v.key for v in self._values.values())
            super(Enum, self).__setattr__('_keyStrings', keyStrings)
        return self._keyStrings

# strangely this is required to fix a crash when referencing pymel from mypy
EnumType = Enum

from . import utilitytypes


class EnumDict(utilitytypes.EquivalencePairs):

    """
    This class provides a dictionary type for storing enumerations.  Keys are string labels, while
    values are enumerated integers.

    To instantiate, pass a sequence of string arguments to the EnumDict() constructor:

        >>> from .enum import EnumDict
        >>> Colours = EnumDict(['red', 'blue', 'green'])
        >>> Weekdays = EnumDict(['mon', 'tue', 'wed', 'thu', 'fri', 'sat', 'sun'])
        >>> sorted(Weekdays.items())
        [('fri', 4), ('mon', 0), ('sat', 5), ('sun', 6), ('thu', 3), ('tue', 1), ('wed', 2)]

    Alternately, a dictionary of label-value pairs can be provided:

        >>> Numbers = EnumDict({'one': 1, 'two': 2, 'hundred' : 100, 'thousand' : 1000 } )

    To convert from one representation to another, just use normal dictionary retrieval, it
    works in either direction:

        >>> Weekdays[4]
        'fri'
        >>> Weekdays['fri']
        4

    If you need a particular representation, but don't know what you're starting from ( for
    example, a value that was passed as an argument ) you can use `EnumDict.key` or
    `EnumDict.value`:

        >>> Weekdays.value(3)
        3
        >>> Weekdays.value('thu')
        3
        >>> Weekdays.key(2)
        'wed'
        >>> Weekdays.key('wed')
        'wed'
    """

    def __init__(self, keys, **kwargs):
        """ Create an enumeration instance """

        if not keys:
            raise EnumEmptyError()

        if isinstance(keys, Mapping):
            items = list(keys.items())
            if isinstance(items[0][0], int):
                byKey = dict((k, v) for v, k in items)
            else:
                byKey = keys
        else:
            byKey = dict((k, v) for v, k in enumerate(keys))
        super(EnumDict, self).__init__(byKey)

#        for key, value in byKey.items():
#            try:
#                super(EnumDict, self).__setattr__(key, value)
#            except TypeError, e:
#                raise EnumBadKeyError(key)
#
#        super(EnumDict, self).__setattr__('_reverse', {})
#        self.update(byKey)
#
#    def __setattr__(self, name, value):
#        raise EnumImmutableError(name)
#
#    def __delattr__(self, name):
#        raise EnumImmutableError(name)
#
#    def __setitem__(self, index, value):
#        raise EnumImmutableError(index)
#
#    def __delitem__(self, index):
#        raise EnumImmutableError(index)

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, super(EnumDict, self).__repr__())

    def value(self, key):
        """
        get an index value from a key. this method always returns an index. if a valid index is passed instead of a key, the index will
        be returned unchanged.  this is useful when you need an index, but are not certain whether you are starting with a key or an index.

            >>> units = EnumDict(['invalid', 'inches', 'feet', 'yards', 'miles', 'millimeters', 'centimeters', 'kilometers', 'meters'])
            >>> units.value('inches')
            1
            >>> units.value(3)
            3
            >>> units.value('hectares')
            Traceback (most recent call last):
              ...
            ValueError: invalid enumerator key: 'hectares'
            >>> units.value(10) #doctest: +ELLIPSIS
            Traceback (most recent call last):
              ...
            ValueError: invalid enumerator value: 10
        """
        if isinstance(key, int):
            # got a potential index : checking if it's valid
            if key in self.values():
                return key
            else:
                raise ValueError("invalid enumerator value: %r" % key)
        else:
            # got a key: retrieving index
            try:
                return dict.__getitem__(self, key)
            except KeyError:
                raise ValueError("invalid enumerator key: %r" % key)

    def key(self, index):
        """
        get a key value from an index. this method always returns a key. if a valid key is passed instead of an index, the key will
        be returned unchanged.  this is useful when you need a key, but are not certain whether you are starting with a key or an index.

            >>> units = EnumDict(['invalid', 'inches', 'feet', 'yards', 'miles', 'millimeters', 'centimeters', 'kilometers', 'meters'])
            >>> units.key(2)
            'feet'
            >>> units.key('inches')
            'inches'
            >>> units.key(10)  #doctest: +ELLIPSIS
            Traceback (most recent call last):
              ...
            ValueError: invalid enumerator value: 10
            >>> units.key('hectares')
            Traceback (most recent call last):
              ...
            ValueError: invalid enumerator key: 'hectares'
        """

        if isinstance(index, int):
            # got an index: retrieving key
            try:
                return self._reverse[index]
            except KeyError:
                raise ValueError("invalid enumerator value: %r" % index)
        else:
            # got a potential key : checking if it's valid
            if index in dict.keys(self):
                return index
            else:
                raise ValueError("invalid enumerator key: %r" % index)

    def values(self):
        "return a list of ordered integer values"
        return sorted(dict.values(self))

    def keys(self):
        "return a list of keys as strings ordered by their enumerator value"
        return [self._reverse[v] for v in self.values()]



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
import sys
import os
import re
import os.path
from . import mellex

try:
    from pymel.util.external.ply import *
except ImportError:
    from ply import *

from pymel.util import unescape
import pymel
import pymel.util as util
import pymel.internal.factories as factories

tokens = mellex.tokens


def p_translation_unit(t):
    '''translation_unit : external_declaration
                        | translation_unit external_declaration'''
    pass
    # print "translation_unit"
    #t[0] = assemble(t, 'p_translation_unit')
    # print '\n'

# external-declaration:


def p_external_declaration(t):
    '''external_declaration : function_definition
                            | group'''
    pass
    # print "external_declaration", t[1]
    #t[0] = assemble(t, 'p_external_declaration')
    # if t.lexer.verbose:
    #    print "external_declaration", t[0]

# function-definition:


def p_function_definition(t):
    '''function_definition :  function_declarator function_specifiers_opt ID LPAREN function_arg_list_opt RPAREN group'''
    #t[0] = assemble(t, 'p_function_definition')

    # add to the ordered list of procs
    t.lexer.proc_list.append(t[3])

    # global proc
    if t[1]:
        # print "adding global"
        t.lexer.global_procs[t[3]] = {'returnType': t[2], 'args': t[5]}
        #t[0] = addHeldComments(t, 'func') + "def %s(%s):\n%s\n" % (t[3], ','.join(t[6]) , entabLines( t[9]) )

    # local proc gets prefixed with underscore
    else:
        # print "adding local"
        t.lexer.local_procs[t[3]] = {'returnType': t[2], 'args': t[5]}
        #t[0] = addHeldComments(t, 'func') + "def _%s(%s):\n%s\n" % (t[3], ','.join(t[6]) , entabLines( t[9]) )


def p_function_declarator(t):
    '''function_declarator : GLOBAL PROC
                           | PROC'''
    # string
    # string[]
    #
    if len(t) == 2:
        # print "local proc"
        t[0] = False
    else:
        # print "global proc"
        t[0] = True


def p_type_specifier(t):
    '''type_specifier : INT
                      | FLOAT
                      | STRING
                      | VECTOR
                      | MATRIX
                      '''
    # print "type_specifier"
    t[0] = t[1]

# function-specifiers


def p_function_specifiers_opt(t):
    '''function_specifiers_opt : type_specifier
                                  | type_specifier LBRACKET RBRACKET
                                  | empty'''
    # string
    # string[]
    #
    # print "function_specifiers_opt"
    if len(t) == 2:
        t[0] = t[1]
    else:
        t[0] = t[1] + '[]'
    #t[0] = assemble(t, 'p_function_specifiers_opt')


def p_function_arg(t):
    '''function_arg : type_specifier VAR
                    | type_specifier VAR LBRACKET RBRACKET'''
    #t[0] = assemble(t, 'p_function_arg')
    if len(t) == 3:
        t[0] = (t[1], t[2])
    else:
        t[0] = (t[1] + '[]', t[2])


def p_function_arg_list(t):
    '''function_arg_list : function_arg
                        | function_arg_list COMMA function_arg'''

    #t[0] = assemble(t, 'p_function_arg_list')
    if len(t) > 2:
        t[0] = t[1] + [t[3]]
    # start a new list
    else:
        t[0] = [t[1]]


def p_function_arg_list_opt(t):
    '''function_arg_list_opt : function_arg_list
                        |  empty'''

    #t[0] = assemble(t, 'p_function_arg_list_opt')
    if not t[1]:
        t[0] = []
    else:
        t[0] = t[1]


def p_declaration_specifiers(t):
    '''declaration_specifiers : type_specifier
                              | GLOBAL type_specifier'''
    # print "declaration_specifiers"
    if len(t) == 2:
        t[0] = (None, t[1])
    else:
        t[0] = (t[1], t[2])


def p_group_list_opt(t):
    '''group_list_opt : group_list
                | empty
                '''
    # print "group_list_opt"
    t[0] = t[1]


def p_group_list(t):
    '''group_list : group_list group
            | group'''
    pass
#    print "group_list"
#    if len(t) == 2:
#        t[0] = [t[1]]
#    else:
#        t[0] = t[1] + [t[2]]


def p_group(t):
    '''group : element
            | LBRACE group_list_opt RBRACE'''
    pass
#    if len(t) == 2:
#        print "group", t[1]
#        t[0] = t[1]
#    else:
#        print "adding brackets", t[2]
#        t[0] = t[2]

# def p_element_list_opt(t):
#    '''element_list_opt : element_list
#                | empty
#                '''
#    print "empty"
#    t[0] = t[1]
#
# def p_element_list(t):
#    '''element_list : element
#                | element_list element
#                '''
#    if len(t) == 2:
#        t[0] = [t[1]]
#    else:
#        t[0] = t[1] + [t[2]]


def p_element(t):
    '''element : declaration_specifiers
            | BREAK
            | CASE
            | CONTINUE
            | DEFAULT
            | DO
            | ELSE
            | FALSE
            | FOR
            | IF
            | IN
            | NO
            | ON
            | OFF
            | RETURN
            | SWITCH
            | TRUE
            | WHILE
            | YES
            | ID
            | VAR
            | ICONST
            | FCONST
            | SCONST
            | PLUS
            | MINUS
            | TIMES
            | DIVIDE
            | MOD
            | NOT
            | CROSS
            | LOR
            | LAND
            | LT
            | LE
            | GT
            | GE
            | EQ
            | NE
            | EQUALS
            | TIMESEQUAL
            | DIVEQUAL
            | MODEQUAL
            | PLUSEQUAL
            | MINUSEQUAL
            | CROSSEQUAL
            | COMPONENT
            | PLUSPLUS
            | MINUSMINUS
            | CONDOP
            | LPAREN
            | RPAREN
            | LBRACKET
            | RBRACKET
            | COMMA
            | SEMI
            | COLON
            | CAPTURE
            | LVEC
            | RVEC
            | COMMENT
            | COMMENT_BLOCK
            | ELLIPSIS

            '''
    # print "element", t[1]
    t[0] = t[1]


def p_empty(t):
    '''empty : '''
    pass
    #t[0] = assemble(t, 'p_empty')

# def p_error(t):
#    print "error"
#    pass



"The primary module for maya commands and node classes"
from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import sys
import pymel as _pymel
_pymel.core = sys.modules[__name__]
import pymel.internal.startup as _startup
import pymel.internal as _internal

# will check for the presence of an initilized Maya / launch it
_startup.mayaInit()

import pymel.internal.factories as _factories
import pymel.internal.cmdcache as _cmdcache
import pymel.internal.pmcmds as _pmcmds
_pmcmds.addAllWrappedCmds()

import pymel.api.plugins as _plugins
from pymel.core.general import *
from pymel.core.context import *
from pymel.core.system import *
from pymel.core.windows import *
from pymel.core.animation import *
from pymel.core.effects import *
from pymel.core.modeling import *
from pymel.core.rendering import *
from pymel.core.language import *
from pymel.core.other import *

from pymel import api
from pymel.core import nodetypes
from pymel.core import datatypes
from pymel.core import uitypes
from pymel import versions

nt = nodetypes
dt = datatypes
ui = uitypes

from future.utils import PY2
if PY2:
    from __builtin__ import str

# This is for backwards incompatibility due to a bug in the way LazyLoadModule
# was used, which made all the uitypes available in this namespace
# We may consider deprecating or removing these at some point, so don't rely on
# these always being there!
_uitypes_names = ['AnimCurveEditor', 'AnimDisplay', 'AttrColorSliderGrp',
    'AttrControlGrp', 'AttrEnumOptionMenu', 'AttrEnumOptionMenuGrp',
    'AttrFieldGrp', 'AttrFieldSliderGrp', 'AttrNavigationControlGrp',
    'AttributeMenu', 'BlendShapeEditor', 'BlendShapePanel', 'Button', 'Canvas',
    'ChannelBox', 'CheckBox', 'CheckBoxGrp', 'ClipEditor',
    'ClipSchedulerOutliner', 'CmdScrollFieldExecuter', 'CmdScrollFieldReporter',
    'CmdShell', 'ColorEditor', 'ColorIndexSliderGrp', 'ColorSliderButtonGrp',
    'ColorSliderGrp', 'ColumnLayout', 'CommandLine', 'ConfirmDialog',
    'DefaultLightListCheckBox', 'DeviceEditor', 'DevicePanel', 'DynPaintEditor',
    'DynRelEdPanel', 'DynRelEditor', 'ExclusiveLightCheckBox', 'FloatField',
    'FloatFieldGrp', 'FloatScrollBar', 'FloatSlider', 'FloatSlider2',
    'FloatSliderButtonGrp', 'FloatSliderGrp', 'FontDialog', 'FrameLayout',
    'GlRenderEditor', 'GradientControl', 'GradientControlNoAttr', 'GridLayout',
    'HardwareRenderPanel', 'HelpLine', 'HyperGraph', 'HyperPanel', 'HyperShade',
    'IconTextButton', 'IconTextCheckBox', 'IconTextRadioButton',
    'IconTextRadioCollection', 'IconTextScrollList', 'IconTextStaticLabel',
    'Image', 'IntField', 'IntFieldGrp', 'IntScrollBar', 'IntSlider',
    'IntSliderGrp', 'KeyframeOutliner', 'KeyframeStats', 'LayerButton',
    'LayoutDialog', 'MenuBarLayout', 'MenuEditor', 'MenuSet', 'MessageLine',
    'ModelEditor', 'ModelPanel', 'NameField', 'NodeIconButton',
    'OutlinerEditor', 'OutlinerPanel', 'PalettePort', 'PaneLayout',
    'ProgressBar', 'PromptDialog', 'RadioButton', 'RadioButtonGrp',
    'RadioCollection', 'RadioMenuItemCollection', 'RenderLayerButton',
    'RenderWindowEditor', 'RowColumnLayout', 'ScriptTable', 'ScriptedPanel',
    'ScriptedPanelType', 'ScrollField', 'ScrollLayout', 'Separator',
    'ShelfButton', 'ShelfLayout', 'ShelfTabLayout', 'ShellField',
    'SymbolButton', 'SymbolCheckBox', 'TabLayout', 'Text', 'TextCurves',
    'TextField', 'TextFieldButtonGrp', 'TextFieldGrp', 'ToolButton',
    'ToolCollection', 'Visor']
if not _factories.building:
    for _uiname in _uitypes_names:
        setattr(nodetypes, _uiname, getattr(uitypes, _uiname))
    del _uiname

# create some aliases for legacy reasons
objectTypeUI = windows.objectTypeUI = ui.objectTypeUI
toQtObject = windows.toQtObject = ui.toQtObject
toQtLayout = windows.toQtLayout = ui.toQtLayout
toQtControl = windows.toQtControl = ui.toQtControl
toQtMenuItem = windows.toQtMenuItem = ui.toQtMenuItem
toQtWindow = windows.toQtWindow = ui.toQtWindow

from . import runtime

import maya.cmds as cmds

# these modules are imported anyway so they should not be a performance hit
import pymel.util as util
import pymel.api as api

_logger = _internal.getLogger(__name__)

#: dictionary of plugins and the nodes and commands they register
_pluginData = {}

_module = sys.modules[__name__]


def _addPluginCommand(pluginName, funcName):
    global _pluginData

    if funcName not in _pluginData[pluginName].setdefault('commands', []):
        _pluginData[pluginName]['commands'].append(funcName)
    _logger.debug("Adding plugin command: {} (plugin: {})".format(funcName,
                                                                  pluginName))
    #_logger.debug("adding new command:", funcName)
    _factories.cmdlist[funcName] = _cmdcache.getCmdInfoBasic(funcName)
    _factories.cmdlist[funcName]['plugin'] = pluginName
    _pmcmds.addWrappedCmd(funcName)
    # FIXME: I think we can call a much simpler factory function, because
    # plugins cannnot opt into the complex mutations that functionFactory can apply
    func = _factories.functionFactory(funcName)
    try:
        if func:
            # FIXME: figure out what to do about moduleCmds. could a plugin function be in moduleCmds???
            coreModule = 'pymel.core.%s' % _cmdcache.getModule(funcName, {})  # _factories.moduleCmds)
            if coreModule in sys.modules:
                setattr(sys.modules[coreModule], funcName, func)
            # Note that we add the function to both a core module (ie,
            # pymel.core.other), the pymel.core itself, and pymel.all; this
            # way, we mirror the behavior of 'normal' functions
            setattr(_module, funcName, func)
            if 'pymel.all' in sys.modules:
                setattr(sys.modules['pymel.all'], funcName, func)
        else:
            _logger.warning("failed to create function")
    except Exception as msg:
        from builtins import str
        _logger.warning("exception: %s" % str(msg))


def _addPluginNode(pluginName, mayaType):
    global _pluginData

    if mayaType not in _pluginData[pluginName].setdefault('dependNodes', []):
        _pluginData[pluginName]['dependNodes'].append(mayaType)
    _logger.debug("Adding plugin node: {} (plugin: {})".format(mayaType,
                                                               pluginName))
    extraAttrs = _plugins.pyNodeMethods.get(pluginName, {}).get(mayaType, {})
    return _factories.addCustomPyNode(nodetypes, mayaType, extraAttrs=extraAttrs)


def _removePluginCommand(pluginName, command):
    global _pluginData

    commands = _pluginData.get(pluginName, {}).get('commands', [])
    if command in commands:
        commands.remove(command)
    try:
        _pmcmds.removeWrappedCmd(command)
        _module.__dict__.pop(command, None)
    except KeyError:
        _logger.warning("Failed to remove %s from module %s" %
                        (command, _module.__name__))


def _removePluginNode(pluginName, node):
    global _pluginData

    nodes = _pluginData.get(pluginName, {}).get('dependNodes', [])
    if node in nodes:
        nodes.remove(node)
    _factories.removePyNode(nodetypes, node)


def _stripPluginExt(pluginName):
    if not pluginName:
        return pluginName
    def stripExt(name, ext):
        if name.lower().endswith(ext):
            name = name[:-len(ext)]
        return name
    for ext in ('.mll', '.so', '.dylib', '.py', '.pyc'):
        pluginName = stripExt(pluginName, ext)
    return pluginName


def _pluginLoaded(*args):
    global _pluginData

    if len(args) > 1:
        # 2009 API callback, the args are ( [ pathToPlugin, pluginName ], clientData )
        pluginName = args[0][1]
    else:
        pluginName = args[0]
    pluginName = _stripPluginExt(pluginName)

    if not pluginName:
        return

    # Check to see if plugin is really loaded
    if not (cmds.pluginInfo(pluginName, query=1, loaded=1)):
        return

    # Make sure there are no registered callbacks for this plug-in. It has been
    # reported that some 3rd party plug-ins will enter here twice, causing a
    # "callback id leak" which potentially leads to a crash. The reported
    # scenario was:
    # - Launching mayapy.exe
    # - Opening a Maya scene having a requires statement (to the plug-in)
    # - The plug-in imports pymel, causing initialization and entering here.
    if (pluginName in _pluginData) and 'callbackId' in _pluginData[pluginName] \
            and _pluginData[pluginName]['callbackId'] != None:
        api.MEventMessage.removeCallback(_pluginData[pluginName]['callbackId'])

    _logger.debug("Plugin loaded: %s", pluginName)
    _pluginData[pluginName] = {}

    # Commands
    commands = _plugins.pluginCommands(pluginName)

    if commands:
        # clear out the command list first
        _pluginData[pluginName]['commands'] = []
        for funcName in commands:
            try:
                _addPluginCommand(pluginName, funcName)
            except Exception as e:
                _logger.error("Error adding command %s from plugin %s - %s" %
                              (funcName, pluginName, e))
                _logger.debug(traceback.format_exc())

    # Nodes
    try:
        mayaTypes = cmds.pluginInfo(pluginName, query=1, dependNode=1)
    except Exception:
        _logger.error("Failed to get depend nodes list from %s", pluginName)
        mayaTypes = None
    #apiEnums = cmds.pluginInfo(pluginName, query=1, dependNodeId=1)
    if mayaTypes:
        def addPluginPyNodes(*args):
            try:
                id = _pluginData[pluginName].get('callbackId')
            except KeyError:
                _logger.warning("could not find callback id!")
            else:
                if id is not None:
                    api.MEventMessage.removeCallback(id)
                    if hasattr(id, 'disown'):
                        id.disown()

            _pluginData[pluginName]['dependNodes'] = []
            allTypes = set(cmds.ls(nodeTypes=1))
            for mayaType in mayaTypes:
                # make sure it's a 'valid' type - some plugins list node types
                # that don't show up in ls(nodeTypes=1), and aren't creatable
                # ...perhaps they're abstract types?
                # Unfortunately, can't check this, as only plugin I know of
                # that has such a node - mayalive, mlConstraint - is only
                # available up to 2009, which has a bug with allNodeTypes...
                # Oddly enough, mlConstraint WILL show up in allTypes here,
                # but not after the plugin is loaded / callback finishes...?
                if mayaType not in allTypes:
                    continue
                _addPluginNode(pluginName, mayaType)

        # Note - in my testing, a single api.MFileIO.isReadingFile() call would
        # also catch opening + referencing operations... but in commit
        # 6e53d7818e9363d55d417c3a80ea7df94c4998ec, a check only against
        # isReadingFile is commented out... so I'm playing it safe, and assuming
        # there are edge cases where isOpeningFile is True but isReadingFile is
        # not

        # Detect if we are currently opening/importing a file and load as a callback versus execute now
        if (api.MFileIO.isReadingFile() or api.MFileIO.isOpeningFile() or
                api.MFileIO.isReferencingFile()):
            if api.MFileIO.isReferencingFile():
                _logger.debug("Installing temporary plugin-loaded nodes callback - PostSceneRead")
                id = api.MEventMessage.addEventCallback('PostSceneRead', addPluginPyNodes)
            elif api.MFileIO.isImportingFile():
                _logger.debug("Installing temporary plugin-loaded nodes callback - SceneImported")
                id = api.MEventMessage.addEventCallback('SceneImported', addPluginPyNodes)
            else:
                # pre-2012 referencing operations will fall into this branch,
                # which will not work (ie, pre-2012, plugins loaded due to
                # reference loads will not trigger adding of that plugin's
                # PyNodes).
                # While this is obviously less than ideal, no 2011 versions were
                # available for testing when I made the fix for 2012+, and we
                # decided that making nothing worse would be better than
                # potentially introducing problems/instabilities (ie, see
                # messages in commits 6e53d7818e9363d55d417c3a80ea7df94c4998ec
                # and 81bc5ee28f1775a680449fec8724e21e703a52b8).
                _logger.debug("Installing temporary plugin-loaded nodes callback - SceneOpened")
                id = api.MEventMessage.addEventCallback('SceneOpened', addPluginPyNodes)
            _pluginData[pluginName]['callbackId'] = id
            # scriptJob not respected in batch mode, had to use api
            #cmds.scriptJob( event=('SceneOpened',doSomethingElse), runOnce=1 )
        else:
            _logger.debug("Running plugin-loaded nodes callback")
            # add the callback id as None, so addPluginPyNodes SHOULD know that
            # SOMETHING is always in _pluginData[pluginName]['callbackId'], and
            # if there isn't, then something is wrong...
            _pluginData[pluginName]['callbackId'] = None
            addPluginPyNodes()


def _pluginUnloaded(*args):
    global _pluginData

    if len(args) > 1:
        # 2009 API callback, the args are
        # ( [ pluginName, pathToPlugin ], clientData )  OR
        # ( [ pathToPlugin ], clientData )
        pluginName = args[0][0]
    else:
        pluginName = args[0]
    pluginName = _stripPluginExt(pluginName)

    _logger.debug("Plugin unloaded: %s" % pluginName)

    try:
        data = _pluginData.pop(pluginName)
    except KeyError:
        pass
    else:
        # Commands
        commands = data.pop('commands', [])
        if commands:
            _logger.debug("Removing plugin commands: {}".format(
                ', '.join(commands)))
            for command in commands:
                _removePluginCommand(pluginName, command)

        # Nodes
        nodes = data.pop('dependNodes', [])
        if nodes:
            _logger.debug("Removing plugin nodes: {}".format(', '.join(nodes)))
            for node in nodes:
                _removePluginNode(pluginName, node)


global _pluginLoadedCB
global _pluginUnloadedCB
_pluginLoadedCB = None
_pluginUnloadedCB = None


def _installCallbacks():
    """install the callbacks that trigger new nodes and commands to be added to pymel when a
    plugin loads.  This is called from pymel.__init__
    """

    global _pluginLoadedCB
    if _pluginLoadedCB is None:
        _pluginLoadedCB = True
        _logger.debug("Adding pluginLoaded callback")
        #_pluginLoadedCB = pluginLoadedCallback(module)

        id = api.MSceneMessage.addStringArrayCallback(api.MSceneMessage.kAfterPluginLoad, _pluginLoaded)
        if hasattr(id, 'disown'):
            id.disown()
    else:
        _logger.debug("PluginLoaded callback already exists")

    global _pluginUnloadedCB
    if _pluginUnloadedCB is None:
        _pluginUnloadedCB = True

        # BUG: autodesk still has not add python callback support, and calling this as MEL is not getting the plugin name passed to it
        # mel.unloadPlugin( addCallback='''python("import pymel; pymel._pluginUnloaded('#1')")''' )

        _logger.debug("Adding pluginUnloaded callback")
        id = api.MSceneMessage.addStringArrayCallback(api.MSceneMessage.kAfterPluginUnload, _pluginUnloaded)
        if hasattr(id, 'disown'):
            id.disown()

    else:
        _logger.debug("PluginUnloaded callback already exists")

    # add commands and nodes for plugins loaded prior to importing pymel
    preLoadedPlugins = cmds.pluginInfo(q=1, listPlugins=1)
    if preLoadedPlugins:
        _logger.info("Updating pymel with pre-loaded plugins: %s" % ', '.join(preLoadedPlugins))
        for plugin in preLoadedPlugins:
            _pluginLoaded(plugin)


if not _factories.building:
    _installCallbacks()

# run userSetup.py / initialize MEL...
# ...userStartup.py / .mel may try to add plugins and then use their commands /
# nodes with pymel... so do the plugin stuff first
_startup.finalize()



"""
Maya-related functions, which are useful to both `api` and `core`, including `mayaInit` which ensures
that maya is initialized in standalone mode.
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from builtins import zip
from builtins import object
import os.path
import sys
import glob
import inspect
import pprint
import maya
import maya.OpenMaya as om
import maya.utils

import pymel.util
import pymel.util.py2to3 as util2to3

from pymel.util import picklezip, shellOutput, subpackages, refreshEnviron, universalmethod
from collections import namedtuple
import pymel.versions as versions
from pymel.mayautils import getUserPrefsDir
from pymel.versions import shortName, installName
from . import plogging

from future.utils import PY2

# There are FOUR different ways maya might be started, all of which are
# subtly different, that need to be considered / tested:
#
# 1) Normal gui
# 2) maya -prompt
# 3) Render
# 4) mayapy (or just straight up python)

_logger = plogging.getLogger(__name__)
try:
    import pickle as pickle
except:
    _logger.warning("using pickle instead of cPickle: load performance will be affected")
    import pickle

#from maya.cmds import encodeString

isInitializing = False
# Setting this to False will make finalize() do nothing
finalizeEnabled = True
_finalizeCalled = False

_mayaExitCallbackId = None
_mayaUninitialized = False
_atExitCallbackInstalled = False


def _moduleJoin(*args):
    """
    Joins with the base pymel directory.
    :rtype: string
    """
    moduleDir = os.path.dirname(os.path.dirname(sys.modules[__name__].__file__))
    return os.path.realpath(os.path.join(moduleDir, *args))


def mayaStartupHasRun():
    """
    Returns True if maya.app.startup has already finished, False otherwise.
    """
    return 'maya.app.startup.gui' in sys.modules or 'maya.app.startup.batch' in sys.modules


def mayaStartupHasStarted():
    """
    Returns True if maya.app.startup has begun running, False otherwise.

    It's possible that maya.app.startup is in the process of running (ie,
    in maya.app.startup.basic, calling executeUserSetup) - unlike mayaStartup,
    this will attempt to detect if this is the case.
    """
    return hasattr(maya, 'stringTable')


def setupFormatting():
    import pprint
    import maya.utils

    def myResultCallback(obj):
        return pprint.pformat(obj)
    maya.utils.formatGuiResult = myResultCallback
    # prevent auto-completion generator from getting confused
    maya.utils.formatGuiResult.__module__ = 'maya.utils'

# def loadDynamicLibs():
#    """
#    due to a bug in maya.app.commands many functions do not return any value the first time they are run,
#    especially in standalone mode.  this function forces the loading of all dynamic libraries, which is
#    a very fast and memory-efficient process, which begs the question: why bother dynamically loading?
#
#    this function can only be run after maya.standalone is initialized
#    """
#
#    commandListPath = os.path.realpath( os.environ[ 'MAYA_LOCATION' ] )
#    commandListPath = os.path.join( commandListPath, libdir, 'commandList' )
#
#    import maya.cmds
#    assert hasattr( maya.cmds, 'dynamicLoad'), "maya.standalone must be initialized before running this function"
#    file = open( commandListPath, 'r' )
#    libraries = set( [ line.split()[1] for line in file] )
#    for library in libraries:
#        try:
#            maya.cmds.dynamicLoad(library)
#        except RuntimeError:
#            _logger.debug("Error dynamically loading maya library: %s" % library)

# Will test initialize maya standalone if necessary (like if scripts are run from an exernal interpeter)
# returns True if Maya is available, False either


def mayaInit(forversion=None):
    global _mayaUninitialized
    _mayaUninitialized = False

    result = _mayaInit(forversion=forversion)

    if om.MGlobal.mayaState() == om.MGlobal.kLibraryApp:  # mayapy only
        if pymel_options['fix_mayapy_segfault']:
            fixMayapySegFault()

    return result


def _mayaInit(forversion=None):
    # type: (Any) -> bool
    """ Try to init Maya standalone module, use when running pymel from an external Python inerpreter,
    it is possible to pass the desired Maya version number to define which Maya to initialize


    Part of the complexity of initializing maya in standalone mode is that maya does not populate os.environ when
    parsing Maya.env.  If we initialize normally, the env's are available via maya (via the shell), but not in python
    via os.environ.

    Note: the following example assumes that MAYA_SCRIPT_PATH is not set in your shell environment prior to launching
    python or mayapy.

    >>> import maya.standalone            #doctest: +SKIP
    >>> maya.standalone.initialize()      #doctest: +SKIP
    >>> import maya.mel as mm             #doctest: +SKIP
    >>> print(mm.eval("getenv MAYA_SCRIPT_PATH"))    #doctest: +SKIP
    /Network/Servers/sv-user.luma-pictures.com/luma .....
    >>> import os                         #doctest: +SKIP
    >>> 'MAYA_SCRIPT_PATH' in os.environ  #doctest: +SKIP
    False

    The solution lies in `refreshEnviron`, which copies the environment from the shell to os.environ after maya.standalone
    initializes.

    Returns
    -------
    bool
        returns True if maya.cmds required initializing ( in other words, we are in a standalone python interpreter )

    """
    _logger.debug("startup.mayaInit: called")
    setupFormatting()

    global isInitializing

    # test that Maya actually is loaded and that commands have been initialized,for the requested version

    aboutExists = False
    try:
        from maya.cmds import about
        aboutExists = True
    except ImportError:
        pass

    if aboutExists and mayaStartupHasStarted():
        # if this succeeded, we're initialized
        _logger.debug("startup.mayaInit: maya already started - exiting")
        isInitializing = False
        return False

    _logger.debug("startup.mayaInit: running")

    # for use with pymel compatible maya package
    oldSkipSetup = os.environ.get('MAYA_SKIP_USERSETUP_PY')
    os.environ['MAYA_SKIP_USERSETUP_PY'] = 'on'
    try:
        if not aboutExists and 'maya.standalone' not in sys.modules:
            try:
                _logger.debug("startup.mayaInit: running standalone.initialize")
                import maya.standalone  # @UnresolvedImport
                maya.standalone.initialize(name="python")
            except ImportError as e:
                raise ImportError("%s: pymel was unable to intialize maya.standalone" % e)

        try:
            from maya.cmds import about
        except Exception:
            _logger.error("maya.standalone was successfully initialized, but pymel failed to import maya.cmds (or it was not populated)")
            raise
    finally:
        if oldSkipSetup is None:
            del os.environ['MAYA_SKIP_USERSETUP_PY']
        else:
            os.environ['MAYA_SKIP_USERSETUP_PY'] = oldSkipSetup

    if not mayaStartupHasRun():
        _logger.debug("running maya.app.startup")
        # If we're in 'maya -prompt' mode, and a plugin loads pymel, then we
        # can have a state where maya.standalone has been initialized, but
        # the python startup code hasn't yet been run...
        if about(batch=True):
            import maya.app.startup.batch
        else:
            import maya.app.startup.gui

    # return True, meaning we had to initialize maya standalone
    isInitializing = True
    return True


def initMEL():
    if 'PYMEL_SKIP_MEL_INIT' in os.environ or pymel_options.get('skip_mel_init', False):
        _logger.info("Skipping MEL initialization")
        return

    _logger.debug("initMEL")
    mayaVersion = versions.installName()
    prefsDir = getUserPrefsDir()
    if prefsDir is None:
        _logger.error("could not initialize user preferences: MAYA_APP_DIR not set")
    elif not os.path.isdir(prefsDir):
        _logger.error("could not initialize user preferences: %s does not exist" % prefsDir)

    # TODO : use cmds.internalVar to get paths
    # got this startup sequence from autodesk support
    startup = [
        #'defaultRunTimeCommands.mel',  # sourced automatically
        # os.path.join( prefsDir, 'userRunTimeCommands.mel'), # sourced automatically
        'createPreferencesOptVars.mel',
        'createGlobalOptVars.mel',
        os.path.join(prefsDir, 'userPrefs.mel') if prefsDir else None,
        'initialStartup.mel',
        #$HOME/Documents/maya/projects/default/workspace.mel
        'initialPlugins.mel',
        #'initialGUI.mel', #GUI
        #'initialLayout.mel', #GUI
        # os.path.join( prefsDir, 'windowPrefs.mel'), #GUI
        # os.path.join( prefsDir, 'menuSetPrefs.mel'), #GUI
        #'hotkeySetup.mel', #GUI
        'namedCommandSetup.mel',
        os.path.join(prefsDir, 'userNamedCommands.mel') if prefsDir else None,
        #'initAfter.mel', #GUI
        os.path.join(prefsDir, 'pluginPrefs.mel') if prefsDir else None
    ]
    if pymel_options.get('skip_initial_plugins', False):
        # initialPlugins.mel is not sourced when running maya -batch, but has been included
        # in the pymel startup sequence since time immemorial. see pymel.conf for more info
        _logger.info("Skipping loading Initial Plugins")
        startup.remove('initialPlugins.mel')

    import maya.mel

    callbacks = om.MCallbackIdArray()
    toDelete = []

    # initialStartup.mel will run a `file -f -new`.  This is dangerous, for
    # obvoius resons, so we disable all file news while we run these...
    def rejectNewCallback(boolPtr_retCode, clientData):
        om.MScriptUtil.setBool(boolPtr_retCode, False)

    callbacks.append(om.MSceneMessage.addCheckCallback(
        om.MSceneMessage.kBeforeNewCheck, rejectNewCallback))

    try:
        # additionally, userPrefs.mel will apparently create a bunch of ikSolver
        # nodes (apparently just to set some global ik prefs?)
        # make sure we clean those up...
        def logIkSolverCreation(nodeObj, clientData):
            toDelete.append(om.MObjectHandle(nodeObj))

        callbacks.append(om.MDGMessage.addNodeAddedCallback(
            logIkSolverCreation, "ikSolver"))

        for f in startup:
            _logger.debug("running: %s" % f)
            if f is not None:
                if os.path.isabs(f) and not os.path.exists(f):
                    _logger.warning("Maya startup file %s does not exist" % f)
                else:
                    if PY2:
                        # need to encode backslashes (used for windows paths)
                        if isinstance(f, unicode):
                            encoding = 'unicode_escape'
                        else:
                            encoding = 'string_escape'
                        f = f.encode(encoding)
                    else:
                        # encoding to unicode_escape should add escape
                        # sequences, but also make sure everything is in basic
                        # ascii - so if we decode utf-8 (or ascii), it should
                        # give us a string which is escaped
                        f = f.encode('unicode_escape').decode('utf-8')
                    maya.mel.eval('source "%s"' % f)

    except Exception as e:
        _logger.error("could not perform Maya initialization sequence: failed "
                      "on %s: %s" % (f, e))
    finally:
        om.MMessage.removeCallbacks(callbacks)

        # clean up the created ik solvers
        dgMod = om.MDGModifier()
        for objHandle in toDelete:
            if objHandle.isValid():
                dgMod.deleteNode(objHandle.object())
        dgMod.doIt()

    try:
        # make sure it exists
        res = maya.mel.eval('whatIs "userSetup.mel"')
        if res != 'Unknown':
            maya.mel.eval('source "userSetup.mel"')
    except RuntimeError:
        pass

    _logger.debug("done running mel files")


def initAE():
    try:
        pkg = __import__('AETemplates')
    except ImportError:
        return False
    except Exception:
        import traceback
        traceback.print_exc()
        return False
    else:
        # import subpackages
        for data in subpackages(pkg):
            pass
    return True


def finalize():
    global finalizeEnabled
    global _finalizeCalled
    if not finalizeEnabled or _finalizeCalled:
        return
    _logger.debug('finalizing')
    # Set this to true HERE, as in running userSetup.py,
    # we could end up in here again, inside the initial finalize...
    _finalizeCalled = True

    global isInitializing
    if isInitializing:
        # this module is not encapsulated into functions, but it should already
        # be imported, so it won't run again
        assert 'maya.app.startup.basic' in sys.modules, \
            "something is very wrong. maya.app.startup.basic should be imported by now"
        import maya.app.startup.basic
        maya.app.startup.basic.executeUserSetup()

    state = om.MGlobal.mayaState()
    if state == om.MGlobal.kLibraryApp:  # mayapy only
        initMEL()
        if pymel_options['fix_linux_mayapy_segfault']:
            fixMayapy2011SegFault()
    elif state == om.MGlobal.kInteractive:
        initAE()


def fixMayapySegFault():
    # first, install a maya exit callback that will let us know if uninitialize
    # has already been called...
    def mayaExitCallback(data):
        global _mayaUninitialized
        _mayaUninitialized = True
        import maya.cmds
        # Turn off undo, because maya seems to have a bug where uninitialize()
        # will crash if undo is on, and certain custom conditions have been
        # created - ie, from the command line:

        # mayapy -c 'import maya.standalone
        # maya.standalone.initialize()
        # import maya.cmds as cmds
        # import maya.OpenMaya as om
        # om.MGlobal.executeCommand("""
        # global proc int _pymel_undoOrRedoAvailable()
        # {
        #     return (isTrue("UndoAvailable") || isTrue("RedoAvailable"));
        # }
        # """, False, False)
        # cmds.condition("UndoOrRedoAvailable", initialize=True,
        #                d=["UndoAvailable", "RedoAvailable"],
        #                s="_pymel_undoOrRedoAvailable")
        # cmds.setAttr("persp.tx", 20)
        # print "uninitializing..."
        # maya.standalone.uninitialize()'; echo $?
        # # Seg fault - 139
        maya.cmds.undoInfo(state=False)

    global _mayaExitCallbackId
    if _mayaExitCallbackId is not None:
        om.MMessage.removeCallback(_mayaExitCallbackId)
        if hasattr(_mayaExitCallbackId, 'disown'):
            _mayaExitCallbackId.disown()

    om.MSceneMessage.addCallback(om.MSceneMessage.kMayaExiting,
                                 mayaExitCallback)

    # Now, install a python atexit handler
    def uninitializeMayaOnPythonExit():
        if not _mayaUninitialized:
            import maya.standalone  # @UnresolvedImport
            maya.standalone.uninitialize()
            # These fixed some hangs on exit (specifically, from pymel unittests)
            # in maya 2016.51 (Ext 2, SP1)
            sys.stdout.flush()
            sys.stderr.flush()
            sys.__stdout__.flush()
            sys.__stderr__.flush()

    global _atExitCallbackInstalled
    if not _atExitCallbackInstalled:
        import atexit
        atexit.register(uninitializeMayaOnPythonExit)
        _atExitCallbackInstalled = True

_fixMayapy2011SegFault_applied = False

# Have all the checks inside here, in case people want to insert this in their
# userSetup... it's currently not always on
def fixMayapy2011SegFault():
    # this bug showed up in 2011 + 2012, was apparently fixed in 2013, then
    # showed up again in 2014...
    import platform
    if platform.system() == 'Linux':
        if om.MGlobal.mayaState() == om.MGlobal.kLibraryApp:  # mayapy only
            global _fixMayapy2011SegFault_applied
            if _fixMayapy2011SegFault_applied:
                return

            _fixMayapy2011SegFault_applied = True

            # In linux maya 2011, once maya has been initialized, if you try
            # to do a 'normal' sys.exit, it will crash with a segmentation
            # fault..
            # do a 'hard' os._exit to avoid this

            # note that, since there is no built-in support to tell from
            # within atexit functions what the exit code is, we cannot
            # guarantee returning the "correct" exit code... for instance,
            # if someone does:
            #    raise SystemExit(300)
            # we will instead return a 'normal' exit code of 0
            # ... but in general, the return code is a LOT more reliable now,
            # since it used to ALWAYS return non-zero...

            import sys
            import atexit

            # First, wrap sys.exit to store the exit code...
            _orig_exit = sys.exit

            # This is just in case anybody else needs to access the
            # original exit function...
            if not hasattr(sys, '_orig_exit'):
                sys._orig_exit = _orig_exit

            def exit(status):
                sys._exit_status = status
                _orig_exit(status)
            sys.exit = exit

            def hardExit():
                # run all the other exit handlers registered with
                # atexit, then hard exit... this is easy, because
                # atexit._run_exitfuncs pops funcs off the stack as it goes...
                # so all we need to do is call it again
                import sys
                atexit._run_exitfuncs()
                try:
                    print("pymel: hard exiting to avoid mayapy crash...")
                except Exception:
                    pass
                import os
                import sys

                exitStatus = getattr(sys, '_exit_status', None)
                if exitStatus is None:
                    last_value = getattr(sys, 'last_value', None)
                    if last_value is not None:
                        if isinstance(last_value, SystemExit):
                            try:
                                exitStatus = last_value.args[0]
                            except Exception:
                                pass
                        if exitStatus is None:
                            exitStatus = 1
                if exitStatus is None:
                    exitStatus = 0
                os._exit(exitStatus)
            atexit.register(hardExit)

# Fix for non US encodings in Maya


def encodeFix():
    if mayaInit():
        from maya.cmds import about

        mayaEncode = about(cs=True)
        pyEncode = sys.getdefaultencoding()     # Encoding tel que defini par sitecustomize
        if mayaEncode != pyEncode:             # s'il faut redefinir l'encoding
            # reload (sys)                       # attention reset aussi sys.stdout et sys.stderr
            # sys.setdefaultencoding(newEncode)
            #del sys.setdefaultencoding
            # print "# Encoding changed from '"+pyEncode+'" to "'+newEncode+"' #"
            if not about(b=True):              # si pas en batch, donc en mode UI, redefinir stdout et stderr avec encoding Maya
                import maya.utils
                try:
                    import maya.app.baseUI
                    import codecs
                    # Replace sys.stdin with a GUI version that will request input from the user
                    sys.stdin = codecs.getreader(mayaEncode)(maya.app.baseUI.StandardInput())
                    # Replace sys.stdout and sys.stderr with versions that can output to Maya's GUI
                    sys.stdout = codecs.getwriter(mayaEncode)(maya.utils.Output())
                    sys.stderr = codecs.getwriter(mayaEncode)(maya.utils.Output(error=1))
                except ImportError:
                    _logger.debug("Unable to import maya.app.baseUI")


#===============================================================================
# Cache utilities
#===============================================================================


def getImportableObject(importableName):
    import importlib
    if '.' in importableName:
        modulename, objName = importableName.rsplit('.', 1)
    else:
        # if no module, it's in builtins
        modulename = 'builtins'
        if PY2:
            modulename = '__builtin__'
        objName = importableName
    moduleobj = importlib.import_module(modulename)
    return getattr(moduleobj, objName)


def getImportableName(obj):
    module = inspect.getmodule(obj)
    import builtins
    if PY2:
        import __builtin__ as builtins
    if module == builtins:
        return obj.__name__
    return '{}.{}'.format(module.__name__, obj.__name__)


def _pickledump(data, filename, protocol=-1):
    with open(filename, mode='wb') as file:
        pickle.dump(data, file, protocol)


def _pickleload(filename):
    with open(filename, mode='rb') as file:
        res = pickle.load(file)
    return res

# NOTE: currently, there is no support for pymel to read or write formats
# other than the current.  This is here for documentation, and in case we want
# to add that ability at some point in the future.

# (1, 0) - initial version, that used "eval" instead of exec - didn't contain
#          explicit version
# (1, 1) - version that uses "exec" - ie, data = [...]; has a version as well
# (1, 2) - like (1, 2), but has potential unicode characters in utf-8 encoding,
#          and builtin types are now encoded as '<type bool>' instead of
#          '<type __builtin__.bool>'
PY_CACHE_FORMAT_VERSION = (1, 2)


# In python3, the PrettyPrinter will wrap long strings to match the width.
# The string wrapping in some cases looks better, but in many cases it
# results in strings with only a few characters per line, ie:
#                                                               foo: "This "
#                                                                    "is "
#                                                                    "bad."
# So, we use a custom PrettyPrinter to override string wrapping (giving results
# like we used to get in python2)

# Thanks to Martijn Pieters:
#   https://stackoverflow.com/a/31485450/920545

class NoStringWrappingPrettyPrinter(pprint.PrettyPrinter):
    def _format(self, object, *args):
        if isinstance(object, str):
            width = self._width
            self._width = sys.maxsize
            try:
                super()._format(object, *args)
            finally:
                self._width = width
        else:
            super()._format(object, *args)

py_pformat = NoStringWrappingPrettyPrinter().pformat

if PY2:
    # just use the normal PrettyPrinter
    py_pformat = pprint.pformat

def _pyformatdump(data):
    strdata = 'version = {!r}\n\ndata = {}'.format(PY_CACHE_FORMAT_VERSION,
                                                   py_pformat(data))
    if PY2:
        if not isinstance(strdata, unicode):
            return strdata
    return strdata.encode('utf-8')

def _pycodeload(code):
    globs = {}
    exec(code, globs)
    # we ignore globs['version'] for now... there for potential future changes
    return globs['data']


def _pydump(data, filename):
    with open(filename, mode='wb') as file:
        file.write(_pyformatdump(data))


def _pyload(filename):
    with open(filename, mode='rb') as file:
        text = file.read()
    return _pycodeload(compile(text, filename, "exec"))


def _getpycbytes(source):
    import py_compile
    import tempfile

    # tried using NamedTemporaryFile, but got perm errors from
    # py_compile.compile
    pyfd, pyPath = tempfile.mkstemp(suffix='.py', prefix="py_cache_temp")
    try:
        with os.fdopen(pyfd, "wb") as pyf:
            pyf.write(source)
        pycfd, pycPath = tempfile.mkstemp(suffix='.pyc',
                                          prefix="pyc_cache_temp")
        os.close(pycfd)
        try:
            py_compile.compile(pyPath, pycPath, doraise=True)
            with open(pycPath, 'rb') as pycf:
                return pycf.read()
        finally:
            os.remove(pycPath)
    finally:
        os.remove(pyPath)


def _pyczipdump(data, filename, pyc=True):
    import zipfile
    inZipPath = os.path.basename(filename)
    if inZipPath.lower().endswith('.zip'):
        inZipPath = inZipPath[:-len('.zip')]

    bytes = _pyformatdump(data)
    if pyc:
        bytes = _getpycbytes(bytes)
    with zipfile.ZipFile(filename, 'w', zipfile.ZIP_DEFLATED) as zf:
        zf.writestr(inZipPath, bytes)


def _pyzipdump(data, filename):
    return _pyczipdump(data, filename, pyc=False)


def _pyzipload(filename):
    import zipimport
    importer = zipimport.zipimporter(filename)
    moduleName = os.path.basename(filename).split('.')[0]
    return _pycodeload(importer.get_code(moduleName))


CacheFormat = namedtuple('CacheFormat', ['ext', 'reader', 'writer'])


class PymelCache(object):
    # override these
    NAME = ''   # ie, 'mayaApi'
    DESC = ''   # ie, 'the API cache' - used in error messages, etc

    FORMATS = [
        CacheFormat('.py', _pyload, _pydump),
        CacheFormat('.pyc.zip', _pyzipload, _pyczipdump),
        CacheFormat('.py.zip', _pyzipload, _pyzipdump),
        CacheFormat('.bin', _pickleload, _pickledump),
        CacheFormat('.zip', picklezip.load, picklezip.dump),
    ]
    EXTENSIONS = {x.ext: x for x in FORMATS}
    DEFAULT_EXT = '.py'

    # whether to add the version to the filename when writing out the cache
    USE_VERSION = True

    _lastReadPath = None
    _lastWritePath = None

    def fromRawData(self, rawData):
        '''If a subclass needs to modify data as it is read from the cache
        on disk, do it here'''
        return rawData

    def toRawData(self, data):
        '''If a subclass needs to modify data before it is written to the cache
        on disk, do it here'''
        if PY2:
            # the written out .py cache will not include 'u' prefixes, which
            # makes it easier to diff to python-3-built caches; initially, just
            # using this is caches where we think it's unlikely to affect client
            # code
            isUnicode = lambda x: isinstance(x, unicode)
            data = pymel.util.deepPatch(data, isUnicode, util2to3.trystr)
        return data

    def read(self, path=None, ext=None, ignoreError=False):
        if path is not None and ext is None:
            ext = os.path.splitext(path)[1]
            if not ext:
                ext = None
        if ext is None:
            formats = self.FORMATS
        else:
            formats = [self.EXTENSIONS[ext]]

        for format in formats:
            # if the user provided an explicit ext (or path with ext), we want
            # to use that EXACTLY (ie, might differ in case) - otherwise, we
            # use the standard ext from format.ext
            if path is not None:
                formatPath = path
            else:
                formatPath = self.path(ext=format.ext)

            if not os.path.isfile(formatPath):
                _logger.debug(self._actionMessage('Unable to open', 'from nonexistant path', formatPath))
                continue

            func = format.reader
            _logger.debug(self._actionMessage('Loading', 'from', formatPath))
            try:
                finalData = self.fromRawData(func(formatPath))
            except Exception as e:
                self._errorMsg('read', 'from', formatPath, e)
                if not ignoreError:
                    raise
            else:
                self._lastReadPath = formatPath
                return finalData

    def write(self, data, path=None, ext=None, ignoreError=False):
        import copy
        # when writing data, we dont' actually want to modify the passed in
        # data, as it may be in use... so we make a deepcopy
        data = copy.deepcopy(data)

        if path is not None and ext is None:
            ext = os.path.splitext(path)[1]
        if not ext:
            ext = self.DEFAULT_EXT
        if not path:
            path = self.path(ext=ext)
        format = self.EXTENSIONS[ext]
        func = format.writer
        _logger.info(self._actionMessage('Saving', 'to', path))

        try:
            func(self.toRawData(data), path)
        except Exception as e:
            self._errorMsg('write', 'to', path, e)
            if not ignoreError:
                raise
        else:
            self._lastWritePath = path

    @universalmethod
    def path(self, version=None, ext=None):
        if ext is None:
            ext = self.DEFAULT_EXT
        if self.USE_VERSION:
            if version is not None:
                short_version = str(version)
            elif hasattr(self, 'version'):
                short_version = str(self.version)
            else:
                short_version = shortName()
        else:
            short_version = ''

        newPath = _moduleJoin('cache', self.NAME + short_version)
        return newPath + ext

    @classmethod
    def allVersions(cls, allowEmpty=False):
        import itertools
        import re

        # unlikely they'll have a path with PLACEHOLDER, but better safe than
        # sorry...
        placeholderBase = 'PLACEHOLDER'
        i = 0
        for i in itertools.count():
            placeholder = placeholderBase + str(i)
            fullPath = cls.path(placeholder)
            if fullPath.count(placeholder) == 1:
                break
        dirname, filePattern = os.path.split(fullPath)
        filePattern = re.escape(filePattern)
        filePattern = filePattern.replace(placeholder, '(.*)')
        filePatternRe = re.compile('^' + filePattern + '$')
        versions = []
        for filename in os.listdir(dirname):
            if not os.path.isfile(os.path.join(dirname, filename)):
                continue
            match = filePatternRe.match(filename)
            if match:
                version = match.group(1)
                if allowEmpty or version != '':
                    versions.append(version)
        return sorted(versions)

    @classmethod
    def _actionMessage(cls, action, direction, location):
        '''_actionMessage('eat', 'at', 'Joes') =>
            "eat cls.DESC at 'Joes'"
        '''
        description = cls.DESC
        if description:
            description = ' ' + description
        return "%s%s %s %r" % (action, description, direction, location)

    @classmethod
    def _errorMsg(cls, action, direction, path, error):
        '''_errorMessage('eat', 'at', 'Joes') =>
            'Unable to eat cls.DESC at Joes: error.msg'
        '''
        actionMsg = cls._actionMessage(action, direction, path)
        _logger.raiseLog(_logger.WARNING,
                         "Unable to %s: %s" % (actionMsg, error))
        import traceback
        _logger.debug(traceback.format_exc())


# Considered using named_tuple, but wanted to make data stored in cache
# have as few dependencies as possible - ie, just a simple tuple
class SubItemCache(PymelCache):

    '''Used to store various maya information

    ie, api / cmd data parsed from docs

    To implement, create a subclass, which overrides at least the NAME, DESC,
    and _CACHE_NAMES attributes, and implements the rebuild method.

    Then to access data, you should initialize an instance, then call build;
    build will load the data from the cache file if possible, or call rebuild
    to build the data from scratch if not.  If the data had to be rebuilt,
    a new file cache will be saved.

    The data may then be accessed through attributes on the instance, with
    the names given in _CACHE_NAMES.

    >>> class NodeCache(SubItemCache):
    ...     NAME = 'mayaNodes'
    ...     DESC = 'the maya nodes cache'
    ...     _CACHE_NAMES = ['nodeTypes']
    ...     AUTO_SAVE = False
    ...     def rebuild(self):
    ...         import maya.cmds
    ...         self.nodeTypes = maya.cmds.allNodeTypes(includeAbstract=True)
    >>> cacheInst = NodeCache()
    >>> cacheInst.build()
    >>> 'polyCube' in cacheInst.nodeTypes
    True
    '''
    # Provides a front end for a pickled file, which should contain a
    # tuple of items; each item in the tuple is associated with a name from
    # _CACHE_NAMES

    # override this with a list of names for the items within the cache
    _CACHE_NAMES = []

    # Set this to the initialization contructor for each cache item;
    # if a given cache name is not present in ITEM_TYPES, DEFAULT_TYPE is
    # used
    # These are the types that the contents will 'appear' to be to the end user
    # (ie, the types returned by contents).
    # Should be constructors which can either take no arguments, or
    # a single argument to initialize an instance.
    ITEM_TYPES = {}
    DEFAULT_TYPE = dict
    AUTO_SAVE = True

    def __init__(self):
        for name in self._CACHE_NAMES:
            self.initVal(name)

    @classmethod
    def cacheNames(cls):
        return tuple(cls._CACHE_NAMES)

    @classmethod
    def itemType(cls, name):
        return cls.ITEM_TYPES.get(name, cls.DEFAULT_TYPE)

    @classmethod
    def itemIndex(cls, name):
        return cls._CACHE_NAMES.index(name)

    def initVal(self, name):
        itemType = self.itemType(name)
        if itemType is None:
            val = None
        else:
            val = itemType()
        setattr(self, name, val)

    def build(self):
        """
        Used to rebuild cache, either by loading from a cache file, or
        rebuilding from scratch.
        """
        data = self.load()
        if data is None:
            self.rebuild()
            if self.AUTO_SAVE:
                self.save()

    # override this...
    def rebuild(self):
        """Rebuild cache from scratch

        Unlike 'build', this does not attempt to load a cache file, but always
        rebuilds it by parsing the docs, etc.
        """
        pass

    def update(self, obj, cacheNames=None):
        '''Update all the various data from the given object, which should
        either be a dictionary, a list or tuple with the right number of items,
        or an object with the caches stored in attributes on it.
        '''
        if cacheNames is None:
            cacheNames = self.cacheNames()

        if isinstance(obj, dict):
            for key, val in obj.items():
                setattr(self, key, val)
        elif isinstance(obj, (list, tuple)):
            if len(obj) != len(cacheNames):
                raise ValueError('length of update object (%d) did not match '
                                 'length of cache names (%d)' %
                                 (len(obj), len(cacheNames)))
            for newVal, name in zip(obj, cacheNames):
                setattr(self, name, newVal)
        else:
            for cacheName in cacheNames:
                setattr(self, cacheName, getattr(obj, cacheName))

    def load(self):
        '''Attempts to load the data from the cache on file.

        If it succeeds, it will update itself, and return the loaded items;
        if it fails, it will return None
        '''
        data = self.read(ignoreError=True)
        if data is not None:
            data = tuple(data)
            self.update(data, cacheNames=self._CACHE_NAMES)
        return data

    def save(self, obj=None, ext=None):
        '''Saves the cache

        Will optionally update the caches from the given object (which may be
        a dictionary, or an object with the caches stored in attributes on it)
        before saving
        '''
        if obj is not None:
            self.update(obj)
        data = self.contents()
        self.write(data, ext=ext)

    # was called 'caches'
    def contents(self):
        return tuple(getattr(self, x) for x in self.cacheNames())

#===============================================================================
# Config stuff
#===============================================================================


def getConfigFile():
    return plogging.getConfigFile()


def parsePymelConfig():
    if PY2:
        import ConfigParser as configparser
    else:
        import configparser

    types = {
        'skip_mel_init': 'boolean',
        'check_attr_before_lock': 'boolean',
        'fix_mayapy_segfault': 'boolean',
        'fix_linux_mayapy_segfault': 'boolean',
    }
    defaults = {
        'skip_mel_init': 'off',
        'check_attr_before_lock': 'off',
        'preferred_python_qt_binding': 'pyqt',
        # want to use the "better" fix_mayapy_segfault now that uninitialize is
        # available
        'fix_mayapy_segfault': 'on',
        'fix_linux_mayapy_segfault': 'off',
        'deleted_pynode_name_access': 'warn_deprecated',
    }

    config = configparser.ConfigParser(defaults)
    config.read(getConfigFile())

    d = {}
    for option in config.options('pymel'):
        getter = getattr(config, 'get' + types.get(option, ''))
        d[option] = getter('pymel', option)

    # just to standardize this setting to be lowercase
    d['preferred_python_qt_binding'] = d['preferred_python_qt_binding'].lower()

    return d

pymel_options = parsePymelConfig()




"""
Contains all context command functions (previously 'ctx').
"""
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import pymel.internal.factories as _factories
if False:
    from maya import cmds
else:
    import pymel.internal.pmcmds as cmds  # type: ignore[no-redef]


# ------ Do not edit below this line --------

alignCtx = _factories.getCmdFunc('alignCtx')

arcLenDimContext = _factories.getCmdFunc('arcLenDimContext')

@_factories.addCmdDocs
def art3dPaintCtx(*args, **kwargs):
    res = cmds.art3dPaintCtx(*args, **kwargs)
    wraps = _factories.simpleCommandWraps['art3dPaintCtx']
    for func, wrapCondition in wraps:
        if wrapCondition.eval(kwargs):
            res = func(res)
            break
    return res

@_factories.addCmdDocs
def artAttrCtx(*args, **kwargs):
    res = cmds.artAttrCtx(*args, **kwargs)
    wraps = _factories.simpleCommandWraps['artAttrCtx']
    for func, wrapCondition in wraps:
        if wrapCondition.eval(kwargs):
            res = func(res)
            break
    return res

artAttrPaintVertexCtx = _factories.getCmdFunc('artAttrPaintVertexCtx')

artAttrSkinPaintCtx = _factories.getCmdFunc('artAttrSkinPaintCtx')

artBaseCtx = _factories.getCmdFunc('artBaseCtx')

artFluidAttrCtx = _factories.getCmdFunc('artFluidAttrCtx')

artPuttyCtx = _factories.getCmdFunc('artPuttyCtx')

artSelectCtx = _factories.getCmdFunc('artSelectCtx')

artSetPaintCtx = _factories.getCmdFunc('artSetPaintCtx')

@_factories.addCmdDocs
def artUserPaintCtx(*args, **kwargs):
    if len(args):
        doPassSelf = kwargs.pop('passSelf', False)
    else:
        doPassSelf = False
    for key in ['cc', 'chunkCommand', 'gac', 'getArrayAttrCommand', 'getSurfaceCommand', 'getValueCommand', 'gsc', 'gvc', 'sac', 'setArrayValueCommand', 'setValueCommand', 'svc']:
        try:
            cb = kwargs[key]
            if callable(cb):
                kwargs[key] = _factories.makeUICallback(cb, args, doPassSelf)
        except KeyError:
            pass
    res = cmds.artUserPaintCtx(*args, **kwargs)
    return res

blendCtx = _factories.getCmdFunc('blendCtx')

boxDollyCtx = _factories.getCmdFunc('boxDollyCtx')

boxZoomCtx = _factories.getCmdFunc('boxZoomCtx')

clipEditorCurrentTimeCtx = _factories.getCmdFunc('clipEditorCurrentTimeCtx')

createNurbsCircleCtx = _factories.getCmdFunc('createNurbsCircleCtx')

createNurbsConeCtx = _factories.getCmdFunc('createNurbsConeCtx')

createNurbsCubeCtx = _factories.getCmdFunc('createNurbsCubeCtx')

createNurbsCylinderCtx = _factories.getCmdFunc('createNurbsCylinderCtx')

createNurbsPlaneCtx = _factories.getCmdFunc('createNurbsPlaneCtx')

createNurbsSphereCtx = _factories.getCmdFunc('createNurbsSphereCtx')

createNurbsSquareCtx = _factories.getCmdFunc('createNurbsSquareCtx')

createNurbsTorusCtx = _factories.getCmdFunc('createNurbsTorusCtx')

createPolyConeCtx = _factories.getCmdFunc('createPolyConeCtx')

createPolyCubeCtx = _factories.getCmdFunc('createPolyCubeCtx')

createPolyCylinderCtx = _factories.getCmdFunc('createPolyCylinderCtx')

createPolyHelixCtx = _factories.getCmdFunc('createPolyHelixCtx')

createPolyPipeCtx = _factories.getCmdFunc('createPolyPipeCtx')

createPolyPlaneCtx = _factories.getCmdFunc('createPolyPlaneCtx')

createPolyPlatonicSolidCtx = _factories.getCmdFunc('createPolyPlatonicSolidCtx')

createPolyPrismCtx = _factories.getCmdFunc('createPolyPrismCtx')

createPolyPyramidCtx = _factories.getCmdFunc('createPolyPyramidCtx')

createPolySoccerBallCtx = _factories.getCmdFunc('createPolySoccerBallCtx')

createPolySphereCtx = _factories.getCmdFunc('createPolySphereCtx')

createPolyTorusCtx = _factories.getCmdFunc('createPolyTorusCtx')

ctxAbort = _factories.getCmdFunc('ctxAbort')

ctxCompletion = _factories.getCmdFunc('ctxCompletion')

ctxData = _factories.getCmdFunc('ctxData')

ctxEditMode = _factories.getCmdFunc('ctxEditMode')

ctxTraverse = _factories.getCmdFunc('ctxTraverse')

currentCtx = _factories.getCmdFunc('currentCtx')

currentTimeCtx = _factories.getCmdFunc('currentTimeCtx')

curveAddPtCtx = _factories.getCmdFunc('curveAddPtCtx')

curveBezierCtx = _factories.getCmdFunc('curveBezierCtx')

curveCVCtx = _factories.getCmdFunc('curveCVCtx')

curveEPCtx = _factories.getCmdFunc('curveEPCtx')

curveEditorCtx = _factories.getCmdFunc('curveEditorCtx')

curveMoveEPCtx = _factories.getCmdFunc('curveMoveEPCtx')

curveSketchCtx = _factories.getCmdFunc('curveSketchCtx')

directKeyCtx = _factories.getCmdFunc('directKeyCtx')

distanceDimContext = _factories.getCmdFunc('distanceDimContext')

dollyCtx = _factories.getCmdFunc('dollyCtx')

dpBirailCtx = _factories.getCmdFunc('dpBirailCtx')

dragAttrContext = _factories.getCmdFunc('dragAttrContext')

@_factories.addCmdDocs
def draggerContext(*args, **kwargs):
    if len(args):
        doPassSelf = kwargs.pop('passSelf', False)
    else:
        doPassSelf = False
    for key in ['dc', 'dragCommand', 'finalize', 'fnz', 'hc', 'holdCommand', 'initialize', 'inz', 'pc', 'ppc', 'prePressCommand', 'pressCommand', 'rc', 'releaseCommand']:
        try:
            cb = kwargs[key]
            if callable(cb):
                kwargs[key] = _factories.makeUICallback(cb, args, doPassSelf)
        except KeyError:
            pass
    res = cmds.draggerContext(*args, **kwargs)
    return res

drawExtrudeFacetCtx = _factories.getCmdFunc('drawExtrudeFacetCtx')

dynPaintCtx = _factories.getCmdFunc('dynPaintCtx')

dynParticleCtx = _factories.getCmdFunc('dynParticleCtx')

dynSelectCtx = _factories.getCmdFunc('dynSelectCtx')

dynWireCtx = _factories.getCmdFunc('dynWireCtx')

filterButterworthCtx = _factories.getCmdFunc('filterButterworthCtx')

filterGaussianCtx = _factories.getCmdFunc('filterGaussianCtx')

filterKeyReducerCtx = _factories.getCmdFunc('filterKeyReducerCtx')

filterPeakRemoverCtx = _factories.getCmdFunc('filterPeakRemoverCtx')

graphDollyCtx = _factories.getCmdFunc('graphDollyCtx')

graphSelectContext = _factories.getCmdFunc('graphSelectContext')

graphTrackCtx = _factories.getCmdFunc('graphTrackCtx')

greasePencilCtx = _factories.getCmdFunc('greasePencilCtx')

hotkeyCtx = _factories.getCmdFunc('hotkeyCtx')

ikHandleCtx = _factories.getCmdFunc('ikHandleCtx')

ikSplineHandleCtx = _factories.getCmdFunc('ikSplineHandleCtx')

insertJointCtx = _factories.getCmdFunc('insertJointCtx')

insertKeyCtx = _factories.getCmdFunc('insertKeyCtx')

jointCtx = _factories.getCmdFunc('jointCtx')

keyframeRegionCurrentTimeCtx = _factories.getCmdFunc('keyframeRegionCurrentTimeCtx')

keyframeRegionDirectKeyCtx = _factories.getCmdFunc('keyframeRegionDirectKeyCtx')

keyframeRegionDollyCtx = _factories.getCmdFunc('keyframeRegionDollyCtx')

keyframeRegionInsertKeyCtx = _factories.getCmdFunc('keyframeRegionInsertKeyCtx')

keyframeRegionMoveKeyCtx = _factories.getCmdFunc('keyframeRegionMoveKeyCtx')

keyframeRegionScaleKeyCtx = _factories.getCmdFunc('keyframeRegionScaleKeyCtx')

keyframeRegionSelectKeyCtx = _factories.getCmdFunc('keyframeRegionSelectKeyCtx')

keyframeRegionSetKeyCtx = _factories.getCmdFunc('keyframeRegionSetKeyCtx')

keyframeRegionTrackCtx = _factories.getCmdFunc('keyframeRegionTrackCtx')

lassoContext = _factories.getCmdFunc('lassoContext')

latticeDeformKeyCtx = _factories.getCmdFunc('latticeDeformKeyCtx')

@_factories.addCmdDocs
def manipMoveContext(*args, **kwargs):
    if len(args):
        doPassSelf = kwargs.pop('passSelf', False)
    else:
        doPassSelf = False
    for key in ['pod', 'postCommand', 'postDragCommand', 'prc', 'prd', 'preCommand', 'preDragCommand', 'psc']:
        try:
            cb = kwargs[key]
            if callable(cb):
                kwargs[key] = _factories.makeUICallback(cb, args, doPassSelf)
        except KeyError:
            pass
    res = cmds.manipMoveContext(*args, **kwargs)
    return res

manipMoveLimitsCtx = _factories.getCmdFunc('manipMoveLimitsCtx')

@_factories.addCmdDocs
def manipRotateContext(*args, **kwargs):
    if len(args):
        doPassSelf = kwargs.pop('passSelf', False)
    else:
        doPassSelf = False
    for key in ['pod', 'postCommand', 'postDragCommand', 'prc', 'prd', 'preCommand', 'preDragCommand', 'psc']:
        try:
            cb = kwargs[key]
            if callable(cb):
                kwargs[key] = _factories.makeUICallback(cb, args, doPassSelf)
        except KeyError:
            pass
    res = cmds.manipRotateContext(*args, **kwargs)
    return res

manipRotateLimitsCtx = _factories.getCmdFunc('manipRotateLimitsCtx')

@_factories.addCmdDocs
def manipScaleContext(*args, **kwargs):
    if len(args):
        doPassSelf = kwargs.pop('passSelf', False)
    else:
        doPassSelf = False
    for key in ['pod', 'postCommand', 'postDragCommand', 'prc', 'prd', 'preCommand', 'preDragCommand', 'psc']:
        try:
            cb = kwargs[key]
            if callable(cb):
                kwargs[key] = _factories.makeUICallback(cb, args, doPassSelf)
        except KeyError:
            pass
    res = cmds.manipScaleContext(*args, **kwargs)
    return res

manipScaleLimitsCtx = _factories.getCmdFunc('manipScaleLimitsCtx')

mateCtx = _factories.getCmdFunc('mateCtx')

modelCurrentTimeCtx = _factories.getCmdFunc('modelCurrentTimeCtx')

modelingToolkitSuperCtx = _factories.getCmdFunc('modelingToolkitSuperCtx')

moveKeyCtx = _factories.getCmdFunc('moveKeyCtx')

mpBirailCtx = _factories.getCmdFunc('mpBirailCtx')

orbitCtx = _factories.getCmdFunc('orbitCtx')

panZoomCtx = _factories.getCmdFunc('panZoomCtx')

paramDimContext = _factories.getCmdFunc('paramDimContext')

polyAppendFacetCtx = _factories.getCmdFunc('polyAppendFacetCtx')

polyCreaseCtx = _factories.getCmdFunc('polyCreaseCtx')

polyCreateFacetCtx = _factories.getCmdFunc('polyCreateFacetCtx')

polyCutCtx = _factories.getCmdFunc('polyCutCtx')

polyCutUVCtx = _factories.getCmdFunc('polyCutUVCtx')

polyMergeEdgeCtx = _factories.getCmdFunc('polyMergeEdgeCtx')

polyMergeFacetCtx = _factories.getCmdFunc('polyMergeFacetCtx')

polyRetopoCtx = _factories.getCmdFunc('polyRetopoCtx')

polySelectCtx = _factories.getCmdFunc('polySelectCtx')

polySelectEditCtx = _factories.getCmdFunc('polySelectEditCtx')

polyShortestPathCtx = _factories.getCmdFunc('polyShortestPathCtx')

polySlideEdgeCtx = _factories.getCmdFunc('polySlideEdgeCtx')

polySplitCtx = _factories.getCmdFunc('polySplitCtx')

polySuperCtx = _factories.getCmdFunc('polySuperCtx')

polyVertexNormalCtx = _factories.getCmdFunc('polyVertexNormalCtx')

projectionContext = _factories.getCmdFunc('projectionContext')

propModCtx = _factories.getCmdFunc('propModCtx')

regionSelectKeyCtx = _factories.getCmdFunc('regionSelectKeyCtx')

renderWindowSelectContext = _factories.getCmdFunc('renderWindowSelectContext')

retimeKeyCtx = _factories.getCmdFunc('retimeKeyCtx')

rollCtx = _factories.getCmdFunc('rollCtx')

roundCRCtx = _factories.getCmdFunc('roundCRCtx')

scaleKeyCtx = _factories.getCmdFunc('scaleKeyCtx')

@_factories.addCmdDocs
def scriptCtx(*args, **kwargs):
    if len(args):
        doPassSelf = kwargs.pop('passSelf', False)
    else:
        doPassSelf = False
    for key in ['fcs', 'finalCommandScript', 'tf', 'toolFinish', 'toolStart', 'ts']:
        try:
            cb = kwargs[key]
            if callable(cb):
                kwargs[key] = _factories.makeUICallback(cb, args, doPassSelf)
        except KeyError:
            pass
    res = cmds.scriptCtx(*args, **kwargs)
    return res

sculptMeshCacheCtx = _factories.getCmdFunc('sculptMeshCacheCtx')

selectContext = _factories.getCmdFunc('selectContext')

selectKeyCtx = _factories.getCmdFunc('selectKeyCtx')

selectKeyframeRegionCtx = _factories.getCmdFunc('selectKeyframeRegionCtx')

setEditCtx = _factories.getCmdFunc('setEditCtx')

setKeyCtx = _factories.getCmdFunc('setKeyCtx')

@_factories.addCmdDocs
def shadingGeometryRelCtx(*args, **kwargs):
    if len(args):
        doPassSelf = kwargs.pop('passSelf', False)
    else:
        doPassSelf = False
    for key in ['ofc', 'offCommand', 'onCommand', 'onc']:
        try:
            cb = kwargs[key]
            if callable(cb):
                kwargs[key] = _factories.makeUICallback(cb, args, doPassSelf)
        except KeyError:
            pass
    res = cmds.shadingGeometryRelCtx(*args, **kwargs)
    return res

@_factories.addCmdDocs
def shadingLightRelCtx(*args, **kwargs):
    if len(args):
        doPassSelf = kwargs.pop('passSelf', False)
    else:
        doPassSelf = False
    for key in ['ofc', 'offCommand', 'onCommand', 'onc']:
        try:
            cb = kwargs[key]
            if callable(cb):
                kwargs[key] = _factories.makeUICallback(cb, args, doPassSelf)
        except KeyError:
            pass
    res = cmds.shadingLightRelCtx(*args, **kwargs)
    return res

@_factories.addCmdDocs
def showManipCtx(*args, **kwargs):
    if len(args):
        doPassSelf = kwargs.pop('passSelf', False)
    else:
        doPassSelf = False
    for key in ['tf', 'toolFinish', 'toolStart', 'ts']:
        try:
            cb = kwargs[key]
            if callable(cb):
                kwargs[key] = _factories.makeUICallback(cb, args, doPassSelf)
        except KeyError:
            pass
    res = cmds.showManipCtx(*args, **kwargs)
    return res

skinBindCtx = _factories.getCmdFunc('skinBindCtx')

snapTogetherCtx = _factories.getCmdFunc('snapTogetherCtx')

snapshotBeadContext = _factories.getCmdFunc('snapshotBeadContext')

snapshotBeadCtx = _factories.getCmdFunc('snapshotBeadCtx')

snapshotModifyKeyCtx = _factories.getCmdFunc('snapshotModifyKeyCtx')

softModContext = _factories.getCmdFunc('softModContext')

softModCtx = _factories.getCmdFunc('softModCtx')

softSelectOptionsCtx = _factories.getCmdFunc('softSelectOptionsCtx')

spBirailCtx = _factories.getCmdFunc('spBirailCtx')

srtContext = _factories.getCmdFunc('srtContext')

stitchSurfaceCtx = _factories.getCmdFunc('stitchSurfaceCtx')

superCtx = _factories.getCmdFunc('superCtx')

targetWeldCtx = _factories.getCmdFunc('targetWeldCtx')

texCutContext = _factories.getCmdFunc('texCutContext')

texLatticeDeformContext = _factories.getCmdFunc('texLatticeDeformContext')

texManipContext = _factories.getCmdFunc('texManipContext')

texMoveContext = _factories.getCmdFunc('texMoveContext')

texMoveUVShellContext = _factories.getCmdFunc('texMoveUVShellContext')

texRotateContext = _factories.getCmdFunc('texRotateContext')

texScaleContext = _factories.getCmdFunc('texScaleContext')

texSculptCacheContext = _factories.getCmdFunc('texSculptCacheContext')

texSelectContext = _factories.getCmdFunc('texSelectContext')

texSelectShortestPathCtx = _factories.getCmdFunc('texSelectShortestPathCtx')

texSmoothContext = _factories.getCmdFunc('texSmoothContext')

texSmudgeUVContext = _factories.getCmdFunc('texSmudgeUVContext')

texTweakUVContext = _factories.getCmdFunc('texTweakUVContext')

texWinToolCtx = _factories.getCmdFunc('texWinToolCtx')

textureLassoContext = _factories.getCmdFunc('textureLassoContext')

texturePlacementContext = _factories.getCmdFunc('texturePlacementContext')

threePointArcCtx = _factories.getCmdFunc('threePointArcCtx')

trackCtx = _factories.getCmdFunc('trackCtx')

trimCtx = _factories.getCmdFunc('trimCtx')

tumbleCtx = _factories.getCmdFunc('tumbleCtx')

twoPointArcCtx = _factories.getCmdFunc('twoPointArcCtx')

@_factories.addCmdDocs
def userCtx(*args, **kwargs):
    if len(args):
        doPassSelf = kwargs.pop('passSelf', False)
    else:
        doPassSelf = False
    for key in ['ec', 'editCommand', 'fc', 'finalCommand']:
        try:
            cb = kwargs[key]
            if callable(cb):
                kwargs[key] = _factories.makeUICallback(cb, args, doPassSelf)
        except KeyError:
            pass
    res = cmds.userCtx(*args, **kwargs)
    return res

view2dToolCtx = _factories.getCmdFunc('view2dToolCtx')

walkCtx = _factories.getCmdFunc('walkCtx')

wireContext = _factories.getCmdFunc('wireContext')

wrinkleContext = _factories.getCmdFunc('wrinkleContext')



from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import zip
from builtins import object
import string
import sys
import textwrap
import inspect
import pymel.util as util


def indent(s, margin):
    return '\n'.join((margin + line) if line else line
                     for line in s.split('\n'))


class DocstringBuilder(object):
    DOC_WIDTH = 120

    def __init__(self, cmdName, indentation='    '):
        import pymel.internal.factories as factories
        factories.loadCmdCache()
        factories.loadCmdDocCache()
        self.cmdName = cmdName
        try:
            self.cmdInfo = factories.cmdlist[cmdName]
        except KeyError:
            raise factories.MelCommandMissingError(cmdName)
        self.indentation = indentation

    def startFlagSection(self):
        raise NotImplementedError

    def indent(self, lines):
        return indent(lines, self.indentation)

    def _addFlag(self, flag, docs):
        # flag type
        try:
            argInfo = docs['args']
        except KeyError:
            raise KeyError("Error retrieving doc args for: "
                           "%r, %r" % (self.cmdName, flag))

        typ = self.getTypeIdentifier(argInfo)

        # flag docstring
        descr = docs.get('docstring', '')

        # flag modes
        tmpmodes = docs.get('modes', [])
        modes = []
        if 'create' in tmpmodes:
            modes.append('create')
        if 'query' in tmpmodes:
            modes.append('query')
        if 'edit' in tmpmodes:
            modes.append('edit')

        return self.addFlag(flag, typ, docs['shortname'], descr, modes)

    def addFlag(self, flag, typ, shortname, descr, modes):
        "Generate docs for a single flag"
        raise NotImplementedError

    def addFooter(self):
        return '\nDerived from mel command `maya.cmds.%s`\n' % (self.cmdName)

    def getTypeIdentifier(self, typ):
        if isinstance(typ, list):
            try:
                typ = [x.__name__ for x in typ]
            except:
                typ = [str(x) for x in typ]
            typ = ', '.join(typ)
        else:
            try:
                typ = typ.__name__
            except:
                pass
        return typ

    def build(self, docstring):
        # type: (str) -> str
        """
        Add the generated docstrings to the existing docstring

        Parameters
        ----------
        docstring : str
            original docstring

        Returns
        -------
        str
        """
        if docstring:
            docstring = inspect.cleandoc(docstring)
            docstring = self.indent(docstring) + '\n\n'

        # insert the description before the existing docstring
        description = '\n'.join(
            textwrap.wrap(self.cmdInfo['description'], width=self.DOC_WIDTH))

        docstring = self.indent(description) + '\n\n' + docstring

        flagsInfo = self.cmdInfo['flags']

        if flagsInfo and not set(flagsInfo.keys()).issubset(['edit', 'query']):

            docstring += self.indent(self.startFlagSection())

            for flag in sorted(flagsInfo.keys()):
                if flag in ['edit', 'query']:
                    continue

                docstring += self.indent(self._addFlag(flag, flagsInfo[flag]))

        docstring += self.indent(self.addFooter())
        return '\n' + docstring


class RstDocstringBuilder(DocstringBuilder):

    """
    Docstring builder that outputs reStructuredText for building HTML docs
    """
    widths = [3, 100, 32, 32]
    altwidths = [widths[0] + widths[1]] + widths[2:]
    rowsep = '+' + '+'.join(['-' * (w - 1) for w in widths]) + '+\n'
    headersep = '+' + '+'.join(['=' * (w - 1) for w in widths]) + '+\n'

    @staticmethod
    def section(title):
        return '.. rubric:: %s:\n' % title

    @staticmethod
    def makerow(items, widths):
        return '|' + '|'.join(
            ' ' + i.ljust(w - 2) for i, w in zip(items, widths)) + '|\n'

    def startFlagSection(self):
        s = self.section('Flags')
        s += '\n' + self.rowsep
        s += self.makerow(
            ['Long Name / Short Name', 'Argument Types', 'Properties'],
            self.altwidths)
        s += self.headersep
        return s

    def addFlag(self, flag, typ, shortname, descr, modes):
        s = ''
        for data in util.izip_longest(
                ['``%s`` / ``%s``' % (flag, shortname)],
                textwrap.wrap('*%s*' % typ, self.widths[2] - 2),
                ['.. image:: /images/%s.gif' % m for m in modes],
                fillvalue=''):
            s += self.makerow(data, self.altwidths)

        # docstring += makerow( ['**%s (%s)**' % (flag, docs['shortname']),
        # '*%s*' % typ, ''], altwidths )
        # for m in modes:
        #    docstring += makerow( ['', '', '.. image:: /images/%s.gif' % m], altwidths )

        s += self.rowsep

        descr_widths = [self.widths[0], sum(self.widths[1:])]
        if descr:
            for line in textwrap.wrap(descr.strip('|'), sum(self.widths[1:]) - 2):
                s += self.makerow(['', line], descr_widths)
                # add some filler at the bottom
                # docstring += makerow(['', '  ..'], descr_widths)
        else:
            s += self.makerow(['', ''], descr_widths)

        # empty row for spacing
        # docstring += rowsep
        # docstring += makerow( ['']*len(widths), widths )
        # closing separator
        s += self.rowsep
        return s

    def addFooter(self):
        footer = super(RstDocstringBuilder, self).addFooter()

        if self.cmdInfo.get('example', None):
            # docstring = ".. |create| image:: /images/create.gif\n.. |edit| image::
            # /images/edit.gif\n.. |query| image:: /images/query.gif\n\n" + docstring
            footer += ('\n\n' + self.section('Example') + '\n::\n' +
                       self.cmdInfo['example'])
        return footer


class NumpyDocstringBuilder(DocstringBuilder):
    DOC_WIDTH = 80

    @staticmethod
    def section(title):
        return title + '\n' + ('-' * len(title)) + '\n'

    def startFlagSection(self):
        return self.section('Parameters')

    def getTypeIdentifier(self, typ):
        if isinstance(typ, list):
            try:
                typ = [x.__name__ for x in typ]
            except:
                typ = [str(x) for x in typ]
            typ = 'Tuple[%s]' % ', '.join(typ)
        else:
            try:
                typ = typ.__name__
            except:
                pass
        return typ

    def addFlag(self, flag, typ, shortname, descr, modes):
        extra = '- modes: %s\n' % ', '.join(modes)
        if shortname:
            extra += '- shortname: %s\n' % shortname
        extra = self.indent(extra)

        descr = '\n'.join(
            ['    ' + x for x in textwrap.wrap(descr, self.DOC_WIDTH)])
        # add trailing newline
        descr = descr + '\n' if descr else ''

        docstring = '%s : %s\n%s%s' % (flag, typ, descr, extra)
        return docstring


class PyDocstringBuilder(DocstringBuilder):

    """
    Docstring builder that generates human-readable docstrings for use with
    the python help() function.
    """
    DOC_WIDTH = 80

    @staticmethod
    def section(title):
        return title + ':\n'

    def startFlagSection(self):
        return self.section('Flags')

    def addFlag(self, flag, typ, shortname, descr, modes):
        descr = '\n'.join(
            ['    ' + x for x in textwrap.wrap(descr, self.DOC_WIDTH)])
        # add trailing newline
        descr = descr + '\n' if descr else ''
        return '- %s %s [%s]\n%s\n' % (
            (flag + ' : ' + shortname).ljust(30),
            ('(' + typ + ')').ljust(15),
            ','.join(modes),
            descr)



# TODO:
# modify this so that it uses MCommandMessage.addCommandOutputFilterCallback,
# and filters messages that we will be converting to python, and then just
# write the converted message ourselves straight to sys.stdout (which has been
# set to a maya.utils.MayaOutput)
# also, set this to just use pymel's pymel.api.plugin "plugin", so no need to
# install a this file as a plugin anywhere...


"""Experimental attempt at wiring mel2py into the script editor.

This is certainly broken, but I'm leaving it in case someone finds it interesting.

to create a window:

window;
    paneLayout -configuration "single";
        pymelScrollFieldReporter;
showWindow;
"""
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import


from builtins import object
import sys
import platform
import maya.OpenMaya as OpenMaya
import maya.OpenMayaMPx as OpenMayaMPx
import maya.OpenMayaUI as OpenMayaUI
import pymel.tools.mel2py as mel2py
from maya.cmds import encodeString

mparser = mel2py.melparse.MelParser()
mparser.build()

kPluginCmdName = "pymelScrollFieldReporter"

'''
# pass to scrollField
-clear(-clr)
-text(-t)    string

# can possibly be implemented
-saveSelection(-sv)    string
-saveSelectionToShelf(-svs)
-textLength(-tl)

# difficult or impossible to implement
-selectAll(-sla)
-select(-sl)    int int
-cutSelection(-ct)
-copySelection(-cp)
-pasteSelection(-pst)
-hasFocus(-hf)
-receiveFocusCommand(-rfc)    string

'''

kClear = ('clear', 'clr')
kText = ('text', 't')
kCmdReporter = ('cmdReporter', 'cr')

kMel = 'mel'
kPython = 'python'

filterFlags = {
    # filters
    'convertToPython': ('ctp', OpenMaya.MSyntax.kBoolean, False),
    'filterSourceType': ('fst', OpenMaya.MSyntax.kString, ''),
    'suppressPrintouts': ('spo', OpenMaya.MSyntax.kBoolean, False),
    'suppressInfo': ('si', OpenMaya.MSyntax.kBoolean, False),
    'suppressWarnings': ('sw', OpenMaya.MSyntax.kBoolean, False),
    'suppressErrors': ('se', OpenMaya.MSyntax.kBoolean, False),
    'suppressResults': ('sr', OpenMaya.MSyntax.kBoolean, False),
    'suppressStackTrace': ('sst', OpenMaya.MSyntax.kBoolean, False)
}

globalFilterFlags = {
    # global
    'echoAllCommands': ('eac', OpenMaya.MSyntax.kBoolean, False),
    'lineNumbers': ('ln', OpenMaya.MSyntax.kBoolean, False),
    'stackTrace': ('st', OpenMaya.MSyntax.kBoolean, False)
}

filterFlagNames = ['',
                   'suppressPrintouts',  # 1
                   'suppressInfo',  # 2
                   'suppressWarnings',  # 3
                   'suppressErrors',  # 4
                   'suppressResults',  # 5
                   'suppressStackTrace']  # 6

messageId = 0
messageIdSet = False
sourceType = kMel  # 'mel' or 'python'

allHistory = []

callbackState = 'normal'

updateCmd = ''

# For XP x64 and Vista, system() returns 'Microsoft'
if platform.system() in ('Windows', 'Microsoft'):
    updateCmd = 'string $fWin=`window -q -frontWindow blah` + "|";string $x;for ($x in `lsUI -controlLayouts -long`) {if (startsWith( $x, $fWin ) ) {break;}};scrollField -e -insertionPosition %(len)d -insertText \"%(text)s\" "%(name)s";setFocus("%(name)s");setFocus($x);'
else:
    updateCmd = 'scrollField -e -insertionPosition %(len)d -insertText \"%(text)s\" "%(name)s";'


class Reporter(object):
    cmdReporter = None

    globalFilters = {
        'echoAllCommands': False,
        'lineNumbers': False,
        'stackTrace': False,
    }

    def __init__(self, *args, **kwargs):
        if not args:
            self.name = kPluginCmdName
        else:
            self.name = args[0]

        # set defaults
        self.filters = {
            # filters
            'convertToPython': False,
            'filterSourceType': '',
            'suppressPrintouts': False,
            'suppressInfo': False,
            'suppressWarnings': False,
            'suppressErrors': False,
            'suppressResults': False,
            'suppressStackTrace': False
        }

        self.filters.update(kwargs)
        self.bufferLength = 0
        global allHistory
        self.history = allHistory[:]
        cmd = 'scrollField -wordWrap false -editable false "%s"' % self.name
        self.name = self.executeCommandResult(cmd)
        self.uiMessageID = OpenMayaUI.MUiMessage.addUiDeletedCallback(self.name, uiDeletedCallback, self.name)
        self.refreshHistory()

    def executeCommandOnIdle(self, cmd):
        global callbackState
        if Reporter.globalFilters['echoAllCommands']:
            callbackState = 'ignoreCommand'
        OpenMaya.MGlobal.executeCommandOnIdle(cmd, False)

    def executeCommand(self, cmd):
        global callbackState
        if Reporter.globalFilters['echoAllCommands']:
            callbackState = 'ignoreCommand'
        OpenMaya.MGlobal.executeCommand(cmd, False, False)

    def executeCommandResult(self, cmd):
        global callbackState
        if Reporter.globalFilters['echoAllCommands']:
            callbackState = 'ignoreCommand'
        result = OpenMaya.MGlobal.executeCommandStringResult(cmd, False, False)
        return result

    def lineFilter(self, messageType, sourceType, nativeMsg, convertedMsg):
        filterSourceType = self.filters['filterSourceType']

        if (not filterSourceType or filterSourceType != sourceType) and not self.filters.get(filterFlagNames[messageType], False):
            if self.filters['convertToPython'] and convertedMsg is not None:
                return convertedMsg
            return nativeMsg

    def refreshHistory(self):
        output = ''
        for line in self.history:
            try:
                output += self.lineFilter(*line)
            except TypeError:
                pass
        self.bufferLength = len(output)
        cmd = 'scrollField -e -text \"%s\" "%s";' % (output, self.name)
        self.executeCommand(cmd)

    def appendHistory(self, line):
        self.history.append(line)
        output = self.lineFilter(*line)

        if output is not None:
            global callbackState

            cmd = updateCmd % {'len': self.bufferLength, 'text': output, 'name': self.name}

            # cmd =  % ( self.bufferLength, output, self.name )
            #cmd = 'scrollField -e -insertionPosition %d -insertText \"%s\" "%s";setFocus("%s")' % ( self.bufferLength, output, self.name, self.name )
            self.bufferLength += len(output)

            # f the line is a syntax error, we have to use OnIdle or maya will crash
            # if line[0] == OpenMaya.MCommandMessage.kError and line[1] == kMel and 'Syntax error' in line[2] : #line[2].endswith( 'Syntax error //\n'):
            if line[1] == kMel and (callbackState == 'syntax_error' or line[0] in [OpenMaya.MCommandMessage.kError, OpenMaya.MCommandMessage.kWarning]):
                self.executeCommandOnIdle(cmd)
            else:
                self.executeCommand(cmd)

            return

    def setFilters(self, **filters):
        self.filters.update(filters)
        self.refreshHistory()

    def setGlobalFilters(self, **filters):
        global cmdReporter

        flags = ''
        for key, value in filters.items():
            if value:
                value = 1
            else:
                value = 0
            flags += '-%s %s ' % (key, value)

        cmd = 'cmdScrollFieldReporter -e %s "%s";' % (flags, Reporter.cmdReporter)

        self.executeCommand(cmd)

        Reporter.globalFilters.update(filters)

    def getGlobalFilter(self, filter):
        return Reporter.globalFilters[filter]

        cmd = 'cmdScrollFieldReporter -q -%s "%s";' % (filter, Reporter.cmdReporter)

        #result = self.executeCommandResult( cmd )
        result = OpenMaya.MGlobal.executeCommandStringResult(cmd, False, False)

        return result

    def addCmdReporter(self, cmdReporter):
        Reporter.cmdReporter = cmdReporter

    def clear(self):
        self.history = []
        cmd = 'scrollField -e -clear "%s";' % (self.name)
        self.executeCommand(cmd)

    def text(self, text):
        cmd = 'scrollField -e -text "%s" "%s";' % (text, self.name)
        self.executeCommand(cmd)


class ReporterDict(dict):

    def __getitem__(self, lookupName):
        lookupBuf = lookupName.split('|')
        for key, val in self.items():
            keyBuf = key.split('|')
            if keyBuf[-1 * len(lookupBuf):] == lookupBuf:
                return val
        raise KeyError  # , str(lookupName)

reporters = ReporterDict({})
#reporters = {}


def removeCallback(id):
    try:
        OpenMaya.MMessage.removeCallback(id)
    except:
        sys.stderr.write("Failed to remove callback\n")
        raise


def createCallback(stringData):
    # global declares module level variables that will be assigned
    global messageIdSet
    global messageId

    try:
        id = OpenMaya.MCommandMessage.addCommandOutputCallback(cmdCallback, stringData)
    except:
        sys.stderr.write("Failed to install callback\n")
        messageIdSet = False
    else:
        messageIdSet = True
    return id


def uiDeletedCallback(name):
    #outputFile = open( '/var/tmp/commandOutput', 'a')
    #outputFile.write( 'before=%s\n' % reporters  )
    # outputFile.close()

    removeCallback(reporters[name].uiMessageID)
    reporters.pop(name)

    #outputFile = open( '/var/tmp/commandOutput', 'a')
    #outputFile.write( 'after=%s\n' % reporters  )
    # outputFile.close()


def cmdCallback(nativeMsg, messageType, data):
    global callbackState

    #outputFile = open( '/var/tmp/commandOutput', 'a')
    #outputFile.write( '============\n%s\n%s %s, length %s \n' % (nativeMsg, messageType, callbackState, len(nativeMsg))  )
    # outputFile.close()

    if callbackState == 'ignoreCommand':
        callbackState = 'ignoreResult'
        return
    elif callbackState == 'ignoreResult':
        callbackState = 'normal'
        return

    global sourceType
    global allHistory

    syntaxError = False
    convertedMsg = None

    # Command History
    if messageType == OpenMaya.MCommandMessage.kHistory:
        callbackState = 'normal'
        # if nativeMsg.rfind(';') == len(nativeMsg)-2 : # and len(nativeMsg) >= 2:
        if nativeMsg.endswith(';\n'):  # and len(nativeMsg) >= 2:
            sourceType = kMel
            try:
                #convertedMsg = mel2py.mel2pyStr( nativeMsg )
                convertedMsg = mparser.parse(nativeMsg)
            except Exception as msg:
                syntaxError = True
                pass
        else:
            sourceType = kPython

    # Display - unaltered strings, such as that printed by the print command
    elif messageType == OpenMaya.MCommandMessage.kDisplay and (nativeMsg.endswith(';\n') or nativeMsg.startswith('//')):
        try:
            #convertedMsg = mel2py.mel2pyStr( nativeMsg )
            convertedMsg = mparser.parse(nativeMsg)
        except Exception as msg:
            pass
    else:
        try:
            nativeMsg = {
                # OpenMaya.MCommandMessage.kDisplay: 'Output',
                OpenMaya.MCommandMessage.kInfo: '',
                OpenMaya.MCommandMessage.kWarning: 'Warning: ',
                OpenMaya.MCommandMessage.kError: 'Error: ',
                OpenMaya.MCommandMessage.kResult: 'Result: '
            }[messageType] + nativeMsg

            if sourceType == kMel:
                convertedMsg = '# %s #\n' % nativeMsg
                nativeMsg = '// %s //\n' % nativeMsg
            else:
                nativeMsg = '# %s #\n' % nativeMsg

        except KeyError:
            pass

    nativeMsg = encodeString(nativeMsg)
    if convertedMsg is not None:
        convertedMsg = encodeString(convertedMsg)

    #outputFile = open( '/var/tmp/commandOutput', 'a')
    #outputFile.write( '---------\n%s %s\n' % ( convertedMsg, sourceType ) )
    # outputFile.close()

    line = [messageType, sourceType, nativeMsg, convertedMsg]

    allHistory.append(line)

    # if messageType == OpenMaya.MCommandMessage.kError : # and 'Syntax error' in nativeMsg:
    #    return

    for reporter in reporters.values():
        reporter.appendHistory(line)

    if syntaxError:
        callbackState = 'syntax_error'

    # elif callbackState == 'syntax_error' and 'Syntax error' in nativeMsg:
    #    callbackState = 'normal'

    #global output
    #output += encodeString( message )

    #cmd = 'global string $gCommandReporter;cmdScrollFieldReporter -edit -text \"%s\" $gCommandReporter;' % output
    #cmd = 'scrollField -e -text \"%s\" %s;\n' % ( output, scrollFieldName )

    #OpenMaya.MGlobal.executeCommand( cmd, False, False )

# command


class scriptedCommand(OpenMayaMPx.MPxCommand):

    def __init__(self):
        OpenMayaMPx.MPxCommand.__init__(self)

    def doIt(self, args):
        #global messageId

        try:
            argData = OpenMaya.MArgDatabase(self.syntax(), args)
        except:
            name = kPluginCmdName
            reporter = Reporter(name)
            reporters[reporter.name] = reporter
            return self.setResult(reporter.name)
        try:
            name = argData.commandArgumentString(0)
        except:
            name = kPluginCmdName

        # QUERY
        if argData.isQuery():
            reporter = reporters[name]
            for key, data in filterFlags.items():
                if argData.isFlagSet(key):
                    self.setResult(reporter.filters[key])
                    return

            for key, data in globalFilterFlags.items():
                if argData.isFlagSet(key):
                    self.setResult(reporter.getGlobalFilter(key))
                    return

            if argData.isFlagSet(kCmdReporter[0]):
                self.setResult(reporter.cmdReporter)

        else:
            filters = {}
            for key, data in filterFlags.items():
                if argData.isFlagSet(key):
                    if data[1] == OpenMaya.MSyntax.kBoolean:
                        filters[key] = argData.flagArgumentBool(key, 0)
                    elif data[1] == OpenMaya.MSyntax.kString:
                        filters[key] = argData.flagArgumentString(key, 0)

            globalFilters = {}
            for key, data in globalFilterFlags.items():
                if argData.isFlagSet(key):
                    if data[1] == OpenMaya.MSyntax.kBoolean:
                        globalFilters[key] = argData.flagArgumentBool(key, 0)

            # EDIT
            if argData.isEdit():
                reporter = reporters[name]
                if filters:
                    reporter.setFilters(**filters)
                if globalFilters:
                    reporter.setGlobalFilters(**globalFilters)

                if argData.isFlagSet(kClear[0]):
                    reporter.clear()
                elif argData.isFlagSet(kText[0]):
                    reporter.text(argData.flagArgumentString(kText[0], 0))
                elif argData.isFlagSet(kCmdReporter[0]):
                    reporter.addCmdReporter(argData.flagArgumentString(kCmdReporter[0], 0))
            # CREATE
            else:
                reporter = Reporter(name, **filters)
                reporters[reporter.name] = reporter

                #outputFile = open( '/var/tmp/commandOutput', 'a')
                #outputFile.write( 'create %s\n' % ( reporters ) )
                # outputFile.close()

                #reporters[name] = reporter
            self.setResult(reporter.name)

        #result = OpenMaya.MGlobal.executeCommandStringResult( cmd, False, False )
        #self.setResult( result )
        #

# Creator


def cmdCreator():
    return OpenMayaMPx.asMPxPtr(scriptedCommand())

# Syntax creator


def syntaxCreator():
    syntax = OpenMaya.MSyntax()
    syntax.addArg(OpenMaya.MSyntax.kString)
    syntax.enableQuery(True)
    syntax.enableEdit(True)
    for flag, data in filterFlags.items():
        syntax.addFlag(data[0], flag, data[1])

    for flag, data in globalFilterFlags.items():
        syntax.addFlag(data[0], flag, data[1])

    syntax.addFlag(kClear[1], kClear[0])
    syntax.addFlag(kText[1], kText[0], OpenMaya.MSyntax.kString)
    syntax.addFlag(kCmdReporter[1], kCmdReporter[0], OpenMaya.MSyntax.kString)

    return syntax

# Initialize the script plug-in


def initializePlugin(mobject):
    mplugin = OpenMayaMPx.MFnPlugin(mobject)

    if OpenMaya.MGlobal.mayaVersion() == '8.5':
        raise NotImplementedError("pymelScrollFieldReporter is only supported in Maya 2008 and later.")
    try:
        mplugin.registerCommand(kPluginCmdName, cmdCreator, syntaxCreator)

        global messageIdSet
        global messageId
        if (messageIdSet):
            print("Message callback already installed")
        else:
            # print "Installing callback message"
            messageId = createCallback('')

    except:
        sys.stderr.write("Failed to register command: %s\n" % kPluginCmdName)
        raise

# Uninitialize the script plug-in


def uninitializePlugin(mobject):
    global messageIdSet
    global messageId
    # Remove the callback
    if (messageIdSet):
        removeCallback(messageId)
    # Remove the plug-in command
    mplugin = OpenMayaMPx.MFnPlugin(mobject)
    try:
        mplugin.deregisterCommand(kPluginCmdName)
    except:
        sys.stderr.write("Failed to unregister command: %s\n" % kPluginCmdName)
        raise
